import json
import logging
import os
from argparse import Namespace
from collections.abc import Collection, Mapping, Callable
from functools import cached_property
from pathlib import Path
from typing import Self, Any

import yaml
from colorama import Fore
from dbt.adapters.utils import classproperty
from dbt.artifacts.schemas.catalog import CatalogArtifact
from dbt.cli.main import dbtRunner
from dbt.config import RuntimeConfig
from dbt.contracts.graph.manifest import Manifest

from dbt_contracts import dbt_cli
from dbt_contracts.contracts import Contract, CONTRACT_MAP, ParentContract, ContractContext
from dbt_contracts.contracts.conditions.properties import PathCondition
from dbt_contracts.contracts.result import Result
from dbt_contracts.contracts.utils import get_absolute_project_path, to_tuple
from dbt_contracts.formatters import ResultsFormatter
from dbt_contracts.formatters.table import TableCellBuilder, GroupedTableFormatter, TableFormatter, TableRowBuilder


def _get_default_table_header(result: Result) -> str:
    path = result.path
    header_path = f"{Fore.LIGHTBLUE_EX}{path}{Fore.RESET}"

    if (properties_path := result.properties_path) and properties_path != path:
        header_path += f" @ {Fore.LIGHTCYAN_EX}{properties_path}{Fore.RESET}"

    return f"{Fore.LIGHTWHITE_EX}{result.result_type.rstrip("s")}s{Fore.RESET}: {header_path}"


DEFAULT_TERMINAL_LOG_BUILDER_CELLS = (
    (
        TableCellBuilder(
            key="result_name", colour=Fore.RED, max_width=50
        ),
        TableCellBuilder(
            key="properties_start_line", prefix="L: ", alignment=">", colour=Fore.LIGHTBLUE_EX, min_width=6, max_width=9
        ),
        TableCellBuilder(
            key=lambda result: result.parent_name if result.has_parent else result.name,
            colour=Fore.CYAN, max_width=40
        ),
        TableCellBuilder(
            key="message", colour=Fore.YELLOW, max_width=60, wrap=True
        ),
    ),
    (
        None,
        TableCellBuilder(
            key="properties_start_col", prefix="P: ", alignment=">", colour=Fore.LIGHTBLUE_EX, min_width=6, max_width=9
        ),
        TableCellBuilder(
            key=lambda result: result.name if result.has_parent else "",
            prefix="> ", colour=Fore.CYAN, max_width=40
        ),
        None,
    ),
)

DEFAULT_TERMINAL_LOG_FORMATTER_TABLE = TableFormatter(
    builder=TableRowBuilder(cells=DEFAULT_TERMINAL_LOG_BUILDER_CELLS, colour=Fore.LIGHTWHITE_EX),
    consistent_widths=True,
)

DEFAULT_TERMINAL_LOG_FORMATTER = GroupedTableFormatter(
    formatter=DEFAULT_TERMINAL_LOG_FORMATTER_TABLE,
    group_key=lambda result: f"{result.result_type}-{result.path}",
    header_key=_get_default_table_header,
    sort_key=[
        "result_type",
        "path",
        lambda result: result.parent_name if result.has_parent else result.name,
        lambda result: result.index or 0,
        lambda result: result.name if result.has_parent else "",
    ],
)


class ContractsRunner:
    """Handles loading config for contracts and their execution."""
    default_config_file_name: str = "contracts"
    default_output_file_name: str = "contracts_results"

    # noinspection PyMethodParameters
    @classproperty
    def output_writers_map(cls) -> Mapping[str, Callable[[Collection[Result], Path], Path]]:
        """A mapping of output types to functions that write the results to a file in that format."""
        return {
            "txt": cls._write_results_as_text,
            "text": cls._write_results_as_text,
            "json": cls._write_results_as_json,
            "jsonl": cls._write_results_as_jsonl,
            "github-annotations": cls._write_results_as_github_annotations,
        }

    @property
    def config(self) -> RuntimeConfig:
        """The dbt runtime config"""
        return self._config

    @cached_property
    def dbt(self) -> dbtRunner:
        """The dbt runner"""
        return dbtRunner()

    @cached_property
    def manifest(self) -> Manifest:
        """The dbt manifest"""
        manifest = dbt_cli.get_manifest(runner=self.dbt, config=self.config, logger=self.logger)
        self.dbt.manifest = manifest
        return manifest

    @cached_property
    def catalog(self) -> CatalogArtifact:
        """The dbt catalog"""
        return dbt_cli.get_catalog(runner=self.dbt, config=self.config, logger=self.logger)

    @property
    def paths(self) -> PathCondition | None:
        """An additional set of paths to filter on when filter contract items."""
        return self._paths

    @paths.setter
    def paths(self, value: str | Path | Collection[str | Path]):
        """Set the path patterns to filter on. Expects a collection regex patterns."""
        paths = []
        for path in to_tuple(value):
            path = Path(path)
            if not (project_root := Path(self.config.project_root)).is_absolute():
                project_root = Path(os.getcwd(), project_root)

            path = get_absolute_project_path(path)
            if path.is_relative_to(project_root):
                paths.append(str(path.relative_to(project_root)))

        self._paths = PathCondition(include=paths) if paths else None

    ################################################################################
    ## Creator methods
    ################################################################################
    @classmethod
    def _resolve_config_path(cls, path: str | Path) -> Path:
        path = Path(path).resolve()
        if path.is_dir():
            path = path.joinpath(cls.default_config_file_name)
        return path

    @classmethod
    def from_config(cls, config: RuntimeConfig) -> Self:
        """
        Set up a new runner from the dbt runtime config with custom args parsed from CLI.

        :param config: The dbt runtime config with args associated.
        :return: The configured runner.
        """
        return cls.from_file(config.args.config, config=config)

    @classmethod
    def from_args(cls, args: Namespace) -> Self:
        """
        Set up a new runner from the args parsed from CLI.

        :param args: The parsed CLI args.
        :return: The configured runner.
        """
        return cls.from_file(args.config, config=dbt_cli.get_config(args))

    @classmethod
    def from_file(cls, path: str | Path, config: RuntimeConfig):
        """
        Set up a new runner from the config in a file at the given `path`.

        :param path: The path to the config file. Must be a path to a valid file.
        :param config: The dbt runtime config.
        :return: The configured runner.
        """
        path = Path(path)
        if path.is_dir():
            path = cls._resolve_config_path(path)
        if not path.suffix:
            path = path.with_suffix(".yml")

        if path.suffix.casefold() in {".yml", ".yaml"}:
            return cls.from_yaml(path, config=config)
        elif path.suffix.casefold() in {".json"}:
            return cls.from_json(path, config=config)

        raise Exception(f"Unrecognised config file extension: {path.suffix}")

    @classmethod
    def from_yaml(cls, path: str | Path, config: RuntimeConfig) -> Self:
        """
        Set up a new runner from the config in a yaml file at the given `path`.

        :param path: The path to the yaml file.
            May either be a path to a yaml file or a path to the directory where the file is located.
            If a directory is given, the default file name will be appended.
        :param config: The dbt runtime config.
        :return: The configured runner.
        """
        with cls._resolve_config_path(path).open("r") as file:
            mapping = yaml.full_load(file)
        return cls.from_dict(mapping, config=config)

    @classmethod
    def from_json(cls, path: str | Path, config: RuntimeConfig) -> Self:
        """
        Set up a new runner from the config in a json file at the given `path`.

        :param path: The path to the json file.
            May either be a path to a json file or a path to the directory where the file is located.
            If a directory is given, the default file name will be appended.
        :param config: The dbt runtime config.
        :return: The configured runner.
        """
        with cls._resolve_config_path(path).open("r") as file:
            mapping = json.load(file)
        return cls.from_dict(mapping, config=config)

    @classmethod
    def from_dict(cls, mapping: Mapping[str, Any], config: RuntimeConfig) -> Self:
        """
        Set up a new runner from the given `config`.

        :param mapping: The config to configure the runner with.
        :param config: The dbt runtime config.
        :return: The configured runner.
        """
        contracts = [
            contract
            for key, contract_configs in mapping.get("contracts", {}).items()
            for conf in contract_configs
            for contract in cls._create_contracts_from_config(key, config=conf)
        ]

        obj = cls(contracts, config=config)
        obj.logger.debug(f"Configured {len(contracts)} sets of contracts from config")
        return obj

    @classmethod
    def _create_contracts_from_config(cls, key: str, config: Mapping[str, Any]) -> list[Contract]:
        key = key.replace(" ", "_").casefold().rstrip("s") + "s"
        if key not in CONTRACT_MAP:
            raise Exception(f"Unrecognised term key: {key}")

        contract = CONTRACT_MAP[key].from_dict(config=config)

        contracts = [contract]
        if isinstance(contract, ParentContract):
            contracts.extend(contract.create_child_contract_from_dict(config))

        return contracts

    def __init__(
            self,
            contracts: Collection[Contract],
            config: RuntimeConfig,
            results_formatter: ResultsFormatter = DEFAULT_TERMINAL_LOG_FORMATTER,
    ):
        self.logger = logging.getLogger(f"{__name__}.{self.__class__.__name__}")

        self._contracts = contracts
        self._config = config
        self._results_formatter = results_formatter

        self._paths: PathCondition | None = None

    ################################################################################
    ## Operations
    ################################################################################
    def validate(self, contract_key: str = None, terms: Collection[str] = ()) -> list[Result]:
        """
        Validate the project against the contracts and log the results.

        :param contract_key: Only process the contract with this key.
        :param terms: Only run the terms with these names.
        :return: The results of the validation.
        """
        contracts = [self._get_contract_by_key(contract_key)] if contract_key is not None else self._contracts
        self._set_artifacts_on_contracts(contracts, force=False)

        results: list[Result] = []
        for contract in contracts:
            contract.validate(terms=terms)
            print()

            if not contract.context.results:
                continue

            # copy needed to assert that the results are logged in tests when mocking
            self.log_results(contract.context.results.copy())
            results.extend(contract.context.results)
            contract.context.results.clear()

        if not results:
            self.logger.info(f"{Fore.LIGHTGREEN_EX}All contracts passed successfully{Fore.RESET}")
            return results

        self.logger.error(f"{Fore.LIGHTRED_EX}Found {len(results)} contract violations{Fore.RESET}")
        return results

    def generate(self, contract_key: str = None) -> dict[Path, int]:
        """
        Generate properties for the contracts and log the results.

        :param contract_key: Only process the contract with this key.
        :return: The paths of the updated/created properties files.
        """
        contracts = [self._get_contract_by_key(contract_key)] if contract_key is not None else self._contracts
        self._set_artifacts_on_contracts(contracts, force=True)

        results: dict[Path, int] = {}
        for contract in contracts:
            contract_results = contract.generate()
            for path, count in contract_results.items():
                results[path] = results.get(path, 0) + count

        if not results:
            self.logger.info(f"{Fore.LIGHTYELLOW_EX}No properties files modified{Fore.RESET}")
            return results

        message = f"Creating/updating {len(results)} properties files"
        self.logger.info(f"{Fore.LIGHTMAGENTA_EX}{message}{Fore.RESET}")

        # properties are updated on the class attribute of ContractContext
        # so all contexts store the same PropertiesIO object
        paths = ContractContext.properties.save(results)
        self._log_generated_paths({path: count for path, count in results.items() if path in paths})

        # clear and refresh the manifest by setting onto each contract again
        self.dbt.manifest = None
        self.__dict__["manifest"] = dbt_cli.get_manifest(
            runner=self.dbt, config=self.config, logger=self.logger, refresh=True
        )
        self.dbt.manifest = self.manifest
        self._set_artifacts_on_contracts(contracts)

        return results

    ################################################################################
    ## Helper functions
    ################################################################################
    def _set_artifacts_on_contracts(self, contracts: Collection[Contract], force: bool = False) -> None:
        for contract in self._contracts:
            contract.manifest = self.manifest
        if force or any(contract.needs_catalog for contract in contracts):
            for contract in self._contracts:
                contract.catalog = self.catalog

        for contract in self._contracts:
            if self.paths is not None and contract.validate_conditions(self.paths):
                if self.paths in contract.conditions:
                    continue
                contract.conditions.append(self.paths)

    def _get_contract_by_key(self, contract_key: str = None) -> Contract:
        for contract in self._contracts:
            if contract.config_key == contract_key:
                return contract

        raise Exception(f"Could not find a configured contract for the key: {contract_key}")

    ################################################################################
    ## Results - logging
    ################################################################################
    def log_results(self, results: Collection[Result]) -> None:
        """
        Log the results of the validation.

        :param results: The results of the validation.
        """
        if not results:
            return

        for line in self._build_results(results).split("\n"):
            self.logger.info(line)
        print()

    def _build_results(self, results: Collection[Result]) -> str:
        self._results_formatter.add_results(results)
        return self._results_formatter.build()

    def _log_generated_paths(self, paths: Mapping[Path, int]) -> None:
        if not paths:
            message = "No files updated/created"
            self.logger.info(f"{Fore.LIGHTYELLOW_EX}{message}{Fore.RESET}")
            return

        message = f"Generated/updated the following number of items across {len(paths)} properties files"
        self.logger.info(f"{Fore.LIGHTGREEN_EX}{message}{Fore.RESET}")

        for path, count in paths.items():
            path = path.relative_to(self.config.project_root)
            self.logger.info(f"  - {Fore.LIGHTBLUE_EX}{path}{Fore.RESET}: {Fore.LIGHTGREEN_EX}{count}{Fore.RESET}")
        print()

    ################################################################################
    ## Results - output
    ################################################################################
    def write_results(self, results: Collection[Result], path: str | Path, output_type: str) -> Path | None:
        """
        Write the given results to an output file with the given `output_type`.

        :param results: The results to write.
        :param path: The path to a directory or file to write to.
        :param output_type: The format to write the file to e.g. 'txt', 'json' etc.
        :return: The path the file was written to.
        """
        if not results:
            return

        if (path := Path(path)).is_dir():
            path = path.joinpath(self.default_output_file_name)
        path.parent.mkdir(parents=True, exist_ok=True)

        output_type = output_type.replace("_", "-")
        output_path = self.output_writers_map.get(output_type)(self, results, path)

        log = f"Wrote {output_type} output to {str(output_path)!r}"
        self.logger.info(f"{Fore.LIGHTBLUE_EX}{log}{Fore.RESET}")
        return output_path

    def _write_results_as_text(self, results: Collection[Result], output_path: Path) -> Path:
        output = self._build_results(results)
        with (path := output_path.with_suffix(".txt")).open("w") as file:
            file.write(output)

        return path

    # noinspection PyMethodMayBeStatic
    def _write_results_as_json(self, results: Collection[Result], output_path: Path) -> Path:
        output = [result.model_dump_json() for result in results]
        with (path := output_path.with_suffix(".json")).open("w") as file:
            json.dump(output, file, indent=2)

        return path

    # noinspection PyMethodMayBeStatic
    def _write_results_as_jsonl(self, results: Collection[Result], output_path: Path) -> Path:
        output = [result.model_dump_json() for result in results]
        with (path := output_path.with_suffix(".json")).open("w") as file:
            for result in output:
                json.dump(result, file)
                file.write("\n")

        return path

    # noinspection PyMethodMayBeStatic
    def _write_results_as_github_annotations(self, results: Collection[Result], output_path: Path) -> Path:
        output = [result.as_github_annotation() for result in results if result.can_format_to_github_annotation]
        with (path := output_path.with_suffix(".json")).open("w") as file:
            json.dump(output, file, indent=2)

        return path



from __future__ import annotations

import logging
from abc import ABCMeta, abstractmethod
from collections import defaultdict
from collections.abc import Collection, Iterable, Mapping, MutableSequence
from functools import cached_property
from pathlib import Path
from typing import Any, Self, Type

from colorama import Fore
from dbt.adapters.utils import classproperty
from dbt.artifacts.resources.v1.components import ColumnInfo
from dbt.artifacts.resources.v1.macro import MacroArgument
from dbt.artifacts.schemas.catalog import CatalogArtifact
from dbt.contracts.graph.manifest import Manifest
from dbt.contracts.graph.nodes import ModelNode, SourceDefinition, Macro

from dbt_contracts.contracts._core import ContractContext, ContractPart
from dbt_contracts.contracts.conditions import ContractCondition, properties as c_properties, source as c_source
from dbt_contracts.contracts.generators import PropertiesGenerator, \
    ParentPropertiesGenerator, ChildPropertiesGenerator, \
    model as g_model, source as g_source, column as g_column
from dbt_contracts.contracts.terms import ContractTerm, ChildContractTerm, \
    properties as t_properties, node as t_node, \
    model as t_model, source as t_source, column as t_column, macro as t_macro
from dbt_contracts.types import ItemT, ParentT, NodeT


class Contract[I: Any, T: ContractTerm, G: PropertiesGenerator | None](metaclass=ABCMeta):
    """
    Composes the various :py:class:`.ContractPart` objects that make a contract for specific types
    of dbt objects within a manifest.

    A Contract gets the dbt objects from the dbt artifacts and passes them to the :py:class:`.ContractPart` objects
    for processing.
    """
    __config_key__: str
    __supported_terms__: tuple[type[T]]
    __supported_conditions__: tuple[type[ContractCondition]]
    __supported_generator__: type[G] | None = None

    @property
    def config_key(self) -> str:
        """The key used to identify this contract in a configuration."""
        return self.__config_key__

    @property
    @abstractmethod
    def items(self) -> Iterable[I]:
        """Get all the items that this contract can process from the manifest."""
        raise NotImplementedError

    @property
    @abstractmethod
    def filtered_items(self) -> Iterable[I]:
        """
        Get all the items that this contract can process from the manifest
        filtered according to the given conditions.
        """
        raise NotImplementedError

    @cached_property
    def context(self) -> ContractContext:
        """Generate a context object from the current loaded dbt artifacts"""
        return ContractContext(manifest=self.manifest, catalog=self.catalog)

    @property
    def needs_manifest(self) -> bool:
        """Do any of the terms in this contract require a manifest to execute"""
        return any(term.needs_manifest for term in self.terms)

    @property
    def needs_catalog(self) -> bool:
        """Do any of the terms in this contract require a catalog to execute"""
        return any(term.needs_catalog for term in self.terms)

    @classmethod
    def from_dict(cls, config: Mapping[str, Any], manifest: Manifest = None, catalog: CatalogArtifact = None) -> Self:
        """
        Create a contract from a given configuration.

        :param config: The configuration to create the contract from.
        :param manifest: The dbt manifest to extract items from.
        :param catalog: The dbt catalog to extract information on database objects from.
        :return: The contract.
        """
        return cls(
            manifest=manifest,
            catalog=catalog,
            conditions=cls._create_conditions(config),
            terms=cls._create_terms(config),
            generator=cls._create_generator(config),
        )

    @classmethod
    def _create_conditions(cls, config: str | Mapping[str, Any]) -> list[ContractCondition]:
        # noinspection PyProtectedMember
        conditions_map = {condition._name(): condition for condition in cls.__supported_conditions__}
        conditions_config = config.get("filter", config.get("conditions", []))
        conditions = tuple(
            cls._create_contract_part_from_dict(conf, part_map=conditions_map) for conf in conditions_config
        )
        return [cond for cond in conditions if cond is not None]

    @classmethod
    def _create_terms(cls, config: str | Mapping[str, Any]) -> list[T]:
        # noinspection PyProtectedMember
        terms_map = {term._name(): term for term in cls.__supported_terms__}
        terms_config = config.get("validations", config.get("terms", []))
        terms = tuple(cls._create_contract_part_from_dict(conf, part_map=terms_map) for conf in terms_config)
        return [term for term in terms if term is not None]

    @classmethod
    def _create_generator(cls, config: str | Mapping[str, Any]) -> G | None:
        if cls.__supported_generator__ is None:
            return

        generator_config = config.get("properties", config.get("generator", {}))
        return cls.__supported_generator__(**generator_config)

    @classmethod
    def _create_contract_part_from_dict[T: ContractPart](
            cls, config: str | Mapping[str, Any], part_map: Mapping[str, Type[T]]
    ) -> T | None:
        if isinstance(config, str):
            part_cls = part_map.get(config, part_map.get(config.rstrip("s")))
            kwargs = {}
        elif isinstance(config, Mapping):
            name, kwargs = next(iter(config.items()))
            part_cls = part_map.get(name, part_map.get(name.rstrip("s")))
        else:
            return

        if part_cls is None:
            return
        if not isinstance(kwargs, Mapping):
            # noinspection PyTypeChecker
            kwargs = {next(iter(part_cls.model_fields)): kwargs}

        return part_cls(**kwargs)

    def __init__(
            self,
            manifest: Manifest = None,
            catalog: CatalogArtifact = None,
            conditions: MutableSequence[ContractCondition] = (),
            terms: MutableSequence[T] = (),
            generator: G | None = None,
    ):
        self.logger = logging.getLogger(f"{__name__}.{self.__class__.__name__}")

        if len(terms) > 0 and not self.validate_terms(terms):
            raise Exception("Unsupported terms for this contract.")
        if len(conditions) > 0 and not self.validate_conditions(conditions):
            raise Exception("Unsupported conditions for this contract.")

        #: The dbt manifest to extract items from
        self.manifest = manifest
        #: The dbt catalog to extract information on database objects from
        self.catalog = catalog

        #: The conditions to apply to items when filtering items to process
        self.conditions = conditions
        #: The terms to apply to items when validating items
        self.terms = terms
        #: The generator to use when creating properties files from database objects
        self.generator = generator

    @classmethod
    def validate_terms(cls, terms: ContractTerm | Collection[ContractTerm]) -> bool:
        """.Validate that all the given ``terms`` are supported by this contract."""
        if isinstance(terms, ContractTerm):
            terms = [terms]

        if not cls.__supported_terms__ and len(terms) > 0:
            return False
        elif cls.__supported_terms__ and len(terms) == 0:
            return False

        return all(term.__class__ in cls.__supported_terms__ for term in terms)

    @classmethod
    def validate_conditions(cls, conditions: ContractCondition | Collection[ContractCondition]) -> bool:
        """.Validate that all the given ``conditions`` are supported by this contract."""
        if isinstance(conditions, ContractCondition):
            conditions = [conditions]

        if not cls.__supported_conditions__ and len(conditions) > 0:
            return False
        elif cls.__supported_conditions__ and len(conditions) == 0:
            return False

        return all(condition.__class__ in cls.__supported_conditions__ for condition in conditions)

    @abstractmethod
    def validate(self, terms: Collection[str] = ()) -> list[I]:
        """
        Validate the terms of this contract against the filtered items.

        :param terms: Only run the terms with these names.
        :return: The valid items.
        """
        raise NotImplementedError

    @abstractmethod
    def generate(self) -> dict[Path, int]:
        """
        Generate the properties files for the items in this contract.

        :return: Map of the paths of properties to save to the number of items updated for that path.
        """
        raise NotImplementedError


class ParentContract[I: ItemT, P: ParentT, G: ParentPropertiesGenerator | None](
    Contract[P, ContractTerm, G], metaclass=ABCMeta
):
    __child_contract__: type[ChildContract[I, P]] | None = None

    # noinspection PyMethodParameters
    @classproperty
    def child_config_key(cls) -> str:
        """The key used to identify the child contract for this contract type."""
        return f"{cls.__config_key__}.{cls.__child_contract__.__config_key__}"

    @property
    def filtered_items(self) -> Iterable[P]:
        for item in self.items:
            if not self.conditions or all(condition.run(item) for condition in self.conditions):
                yield item

    def create_child_contract[GC: ChildPropertiesGenerator | None](
            self,
            conditions: MutableSequence[ContractCondition] = (),
            terms: MutableSequence[ChildContractTerm] = (),
            generator: GC = None,
    ) -> ChildContract[I, P, GC] | None:
        """Create a child contract from this parent contract."""
        if self.__child_contract__ is None:
            return
        return self.__child_contract__(parent=self, conditions=conditions, terms=terms, generator=generator)

    def create_child_contract_from_dict(self, config: Mapping[str, Any]) -> list[ChildContract[I, P]]:
        """Create a child contract from this parent contract."""
        if self.__child_contract__ is None:
            return []
        if (config := config.get(self.__child_contract__.__config_key__)) is None:
            return []

        if isinstance(config, Mapping):
            config = [config]

        contracts = [
            self.__child_contract__.from_dict(config=conf, manifest=self.manifest, catalog=self.catalog)
            for conf in config
        ]
        for contract in contracts:
            contract.parent = self

        return contracts

    # noinspection PyUnresolvedReferences
    def validate(self, terms: Collection[str] = ()) -> list[P]:
        if not self.terms:
            return list(self.filtered_items)

        run_terms = [term for term in self.terms if term.name in terms] if terms else self.terms

        key = f"{self.config_key.rstrip("s")}s"
        message = f"Validating {len(tuple(self.filtered_items))} {key} against {len(run_terms)} terms"
        self.logger.info(f"{Fore.LIGHTMAGENTA_EX}{message}{Fore.RESET}")

        return [
            item for item in self.filtered_items
            if all(term.run(item, context=self.context) for term in run_terms)
        ]

    def generate(self) -> dict[Path, int]:
        if self.generator is None:
            return {}

        message = f"Generating properties for {len(tuple(self.filtered_items))} {self.config_key.rstrip("s")}s"
        self.logger.info(f"{Fore.LIGHTMAGENTA_EX}{message}{Fore.RESET}")

        paths = defaultdict[Path, int](int)
        for item in self.filtered_items:
            modified = self.generator.merge(item, context=self.context)
            if not modified:
                continue

            self.generator.update(item, context=self.context)

            try:
                path = self.context.properties.get_path(item, to_absolute=True)
            except FileNotFoundError:
                path = self.generator.generate_properties_path(item)

            paths[path] += 1

        return dict(paths)


class ChildContract[I: ItemT, P: ParentT, G: ChildPropertiesGenerator | None](
    Contract[tuple[I, P], ChildContractTerm, G], metaclass=ABCMeta
):
    """
    Composes the terms and conditions that make a contract for specific types of dbt child objects within a manifest.
    """
    @property
    def config_key(self) -> str:
        return self.parent.child_config_key

    @property
    @abstractmethod
    def items(self) -> Iterable[tuple[I, P]]:
        """Get all the items that this contract can process from the manifest."""
        raise NotImplementedError

    @property
    def filtered_items(self) -> Iterable[tuple[I, P]]:
        for item, parent in self.items:
            if not self.conditions or all(condition.run(item) for condition in self.conditions):
                yield item, parent

    def __init__(
            self,
            manifest: Manifest = None,
            catalog: CatalogArtifact = None,
            conditions: MutableSequence[ContractCondition] = (),
            terms: MutableSequence[ChildContractTerm] = (),
            generator: G | None = None,
            parent: ParentContract[I, P, ParentPropertiesGenerator | None] = None,
    ):
        super().__init__(
            manifest=manifest or (parent.manifest if parent is not None else None),
            catalog=catalog or (parent.catalog if parent is not None else None),
            conditions=conditions,
            terms=terms,
            generator=generator,
        )

        #: The contract object representing the parent contract for this child contract.
        self.parent = parent

    # noinspection PyUnresolvedReferences
    def validate(self, terms: Collection[str] = ()) -> list[tuple[I, P]]:
        if not self.terms:
            return list(self.filtered_items)

        run_terms = [term for term in self.terms if term.name in terms] if terms else self.terms

        key = f"{self.parent.config_key.rstrip('s')} {self.__config_key__.rstrip("s")}s"
        message = f"Validating {len(tuple(self.filtered_items))} {key} against {len(run_terms)} terms"
        self.logger.info(f"{Fore.LIGHTMAGENTA_EX}{message}{Fore.RESET}")

        return [
            (item, parent) for item, parent in self.filtered_items
            if all(term.run(item, parent=parent, context=self.context) for term in run_terms)
        ]

    def generate(self) -> dict[Path, int]:
        if self.generator is None or self.parent.generator is None:
            return {}

        key = f"{self.parent.config_key.rstrip('s')} {self.__config_key__.rstrip("s")}s"
        message = f"Generating properties for {len(tuple(self.filtered_items))} {key}"
        self.logger.info(f"{Fore.LIGHTMAGENTA_EX}{message}{Fore.RESET}")

        paths = defaultdict[Path, int](int)
        for item, parent in self.filtered_items:
            modified = self.generator.merge(item, context=self.context, parent=parent)
            if not modified:
                continue

            self.parent.generator.update(parent, context=self.context)

            try:
                path = self.context.properties.get_path(parent, to_absolute=True)
            except FileNotFoundError:
                path = self.parent.generator.generate_properties_path(parent)
            paths[path] += 1

        return dict(paths)


class ColumnContract[T: NodeT](ChildContract[ColumnInfo, T, g_column.ColumnPropertiesGenerator]):

    __config_key__ = "columns"

    __supported_terms__ = (
        t_properties.HasDescription,
        t_properties.HasRequiredTags,
        t_properties.HasAllowedTags,
        t_properties.HasRequiredMetaKeys,
        t_properties.HasAllowedMetaKeys,
        t_properties.HasAllowedMetaValues,
        t_column.Exists,
        t_column.HasTests,
        t_column.HasExpectedName,
        t_column.HasDataType,
        t_column.HasMatchingDescription,
        t_column.HasMatchingDataType,
        t_column.HasMatchingIndex,
    )
    __supported_conditions__ = (
        c_properties.NameCondition,
        c_properties.TagCondition,
        c_properties.MetaCondition,
    )
    __supported_generator__ = g_column.ColumnPropertiesGenerator

    @property
    def items(self) -> Iterable[tuple[ColumnInfo, T]]:
        return (
            (col, parent) for parent in self.parent.filtered_items for col in parent.columns.values()
        )


class MacroArgumentContract(ChildContract[MacroArgument, Macro, None]):

    __config_key__ = "arguments"

    __supported_terms__ = (
        t_properties.HasDescription,
        t_macro.HasType,
    )
    __supported_conditions__ = (
        c_properties.NameCondition,
    )

    @property
    def items(self) -> Iterable[tuple[MacroArgument, Macro]]:
        return (
            (arg, mac) for mac in self.parent.filtered_items
            if mac.package_name == self.manifest.metadata.project_name
            for arg in mac.arguments
        )


class ModelContract(ParentContract[ColumnInfo, ModelNode, g_model.ModelPropertiesGenerator]):

    __config_key__ = "models"
    __child_contract__ = ColumnContract

    __supported_terms__ = (
        t_properties.HasProperties,
        t_properties.HasDescription,
        t_properties.HasRequiredTags,
        t_properties.HasAllowedTags,
        t_properties.HasRequiredMetaKeys,
        t_properties.HasAllowedMetaKeys,
        t_properties.HasAllowedMetaValues,
        t_node.Exists,
        t_node.HasTests,
        t_node.HasAllColumns,
        t_node.HasExpectedColumns,
        t_node.HasMatchingDescription,
        t_node.HasContract,
        t_node.HasValidRefDependencies,
        t_node.HasValidSourceDependencies,
        t_node.HasValidMacroDependencies,
        t_node.HasNoFinalSemicolon,
        t_node.HasNoHardcodedRefs,
        t_model.HasConstraints,
    )
    __supported_conditions__ = (
        c_properties.NameCondition,
        c_properties.PathCondition,
        c_properties.TagCondition,
        c_properties.MetaCondition,
        c_properties.IsMaterializedCondition,
    )
    __supported_generator__ = g_model.ModelPropertiesGenerator

    @property
    def items(self) -> Iterable[ModelNode]:
        return (
            node for node in self.manifest.nodes.values()
            if isinstance(node, ModelNode) and node.package_name == self.manifest.metadata.project_name
        )


class SourceContract(ParentContract[ColumnInfo, SourceDefinition, g_source.SourcePropertiesGenerator]):

    __config_key__ = "sources"
    __child_contract__ = ColumnContract

    __supported_terms__ = (
        t_properties.HasProperties,
        t_properties.HasDescription,
        t_properties.HasRequiredTags,
        t_properties.HasAllowedTags,
        t_properties.HasRequiredMetaKeys,
        t_properties.HasAllowedMetaKeys,
        t_properties.HasAllowedMetaValues,
        t_node.Exists,
        t_node.HasTests,
        t_node.HasAllColumns,
        t_node.HasExpectedColumns,
        t_node.HasMatchingDescription,
        t_source.HasLoader,
        t_source.HasFreshness,
        t_source.HasDownstreamDependencies,
    )
    __supported_conditions__ = (
        c_properties.NameCondition,
        c_properties.PathCondition,
        c_properties.TagCondition,
        c_properties.MetaCondition,
        c_source.IsEnabledCondition,
    )
    __supported_generator__ = g_source.SourcePropertiesGenerator

    @property
    def items(self) -> Iterable[SourceDefinition]:
        return (
            source for source in self.manifest.sources.values()
            if source.package_name == self.manifest.metadata.project_name
        )


class MacroContract(ParentContract[MacroArgument, Macro, None]):

    __config_key__ = "macros"
    __child_contract__ = MacroArgumentContract

    __supported_terms__ = (
        t_properties.HasProperties,
        t_properties.HasDescription,
    )
    __supported_conditions__ = (
        c_properties.NameCondition,
        c_properties.PathCondition,
    )

    @property
    def items(self) -> Iterable[Macro]:
        return (
            macro for macro in self.manifest.macros.values()
            if macro.package_name == self.manifest.metadata.project_name
        )


CONTRACT_CLASSES: tuple[type[ParentContract], ...] = (ModelContract, SourceContract, MacroContract)
CONTRACT_MAP = {contract.__config_key__: contract for contract in CONTRACT_CLASSES}



from dbt.contracts.graph.nodes import SourceDefinition

from dbt_contracts.contracts.conditions import ContractCondition


class IsEnabledCondition(ContractCondition[SourceDefinition]):
    """Filter sources taking only those which are enabled."""
    def run(self, item: SourceDefinition) -> bool:
        return item.config.enabled



from random import sample

import pytest
from dbt.contracts.graph.nodes import SourceDefinition

from dbt_contracts.contracts import ContractContext
from dbt_contracts.contracts.generators.source import SourcePropertiesGenerator
from tests.contracts.generators.test_node import NodePropertiesGeneratorTester


class TestSourcePropertiesGenerator(NodePropertiesGeneratorTester[SourceDefinition]):
    @pytest.fixture
    def generator(self) -> SourcePropertiesGenerator:
        return SourcePropertiesGenerator()

    @pytest.fixture
    def item(self, source: SourceDefinition) -> SourceDefinition:
        return source

    def test_generate_source_properties(self, generator: SourcePropertiesGenerator, item: SourceDefinition):
        table = generator._generate_source_properties(item)
        assert all(val for val in table.values())

    def test_generate_table_properties(self, generator: SourcePropertiesGenerator, item: SourceDefinition):
        table = generator._generate_table_properties(item)
        assert all(val for val in table.values())

    def test_generate_properties(self, generator: SourcePropertiesGenerator, item: SourceDefinition):
        properties = generator._generate_properties(item)
        assert item.resource_type.pluralize() in properties

        columns = list(map(generator._generate_column_properties, item.columns.values()))
        table = generator._generate_table_properties(item) | {"columns": columns}
        source = generator._generate_source_properties(item) | {"tables": [table]}

        assert source in properties[item.resource_type.pluralize()]
        for key, val in generator._properties_defaults.items():
            assert properties[key] == val

    def test_update_existing_properties_with_empty_properties(
            self, generator: SourcePropertiesGenerator, item: SourceDefinition, context: ContractContext
    ):
        key = item.resource_type.pluralize()
        properties = {}
        expected_source = generator._generate_properties(item)[key][0]

        assert generator._update_existing_properties(item, properties=properties) is properties
        assert len(properties[key]) == 1
        assert expected_source in properties[key]

    # noinspection PyTestUnpassedFixture
    def test_update_existing_properties_with_new_source(
            self,
            generator: SourcePropertiesGenerator,
            item: SourceDefinition,
            sources: list[SourceDefinition],
            context: ContractContext
    ):
        key = item.resource_type.pluralize()
        sources = sample([source for source in sources if source.name != item.name], k=5)
        properties = {key: [sources[key][0] for sources in map(generator._generate_properties, sources)]}
        assert not any(source["name"] == item.source_name for source in properties[key])

        original_sources_count = len(properties[key])
        expected_source = generator._generate_properties(item)[key][0]

        generator._update_existing_properties(item, properties=properties)
        assert len(properties[key]) == original_sources_count + 1
        assert expected_source in properties[key]

    # noinspection PyTestUnpassedFixture
    def test_update_existing_properties_with_new_table(
            self,
            generator: SourcePropertiesGenerator,
            item: SourceDefinition,
            sources: list[SourceDefinition],
            context: ContractContext
    ):
        key = item.resource_type.pluralize()
        sources = sample([source for source in sources if source.name != item.name], k=5)
        properties = {key: [sources[key][0] for sources in map(generator._generate_properties, sources)]}
        properties[key].append(generator._generate_properties(item)[key][0])
        assert sum(source["name"] == item.source_name for source in properties[key]) == 1

        original_sources_count = len(properties[key])
        expected_columns = list(map(generator._generate_column_properties, item.columns.values()))
        expected_table = generator._generate_table_properties(item) | {"columns": expected_columns}

        generator._update_existing_properties(item, properties=properties)
        assert len(properties[key]) == original_sources_count

        actual_sources = [source for source in properties[key] if source["name"] == item.source_name]
        assert len(actual_sources) == 1
        assert expected_table in actual_sources[0]["tables"]

    # noinspection PyTestUnpassedFixture
    def test_update_existing_properties_with_existing_table(
            self,
            generator: SourcePropertiesGenerator,
            item: SourceDefinition,
            sources: list[SourceDefinition],
            context: ContractContext
    ):
        key = item.resource_type.pluralize()
        sources = sample([source for source in sources if source.name != item.name], k=5)
        source = generator._generate_properties(item)[key][0]
        properties = {key: [sources[key][0] for sources in map(generator._generate_properties, sources)] + [source]}
        assert sum(source["name"] == item.source_name for source in properties[key]) == 1
        assert len(source["tables"]) == 1

        # should update the description in the properties
        original_sources_count = len(properties[key])
        item.description = "a brand new description"

        generator._update_existing_properties(item, properties=properties)
        assert len(properties[key]) == original_sources_count

        actual_sources = [source for source in properties[key] if source["name"] == item.source_name]
        assert len(actual_sources) == 1

        actual_tables = [prop for prop in actual_sources[0]["tables"] if prop["name"] == item.name]
        assert len(actual_tables) == 1
        assert actual_tables[0]["description"] == item.description



from __future__ import annotations

import re
from abc import ABCMeta, abstractmethod
from collections.abc import Sequence
from pathlib import Path
from typing import Literal, Annotated, Any

from dbt.artifacts.resources import BaseResource
from dbt.artifacts.resources.v1.components import ParsedResource
from dbt.flags import get_flags
from dbt_common.dataclass_schema import dbtClassMixin
from pydantic import Field, BeforeValidator

from dbt_contracts.contracts._core import ContractContext, ContractPart
from dbt_contracts.contracts.utils import to_tuple
from dbt_contracts.types import ItemT, PropertiesT

CORE_FIELDS = Literal["description"]


class PropertyGenerator[S: ItemT, T: dbtClassMixin](ContractPart, metaclass=ABCMeta):
    """
    A part of a contract meant to generate a property for a resource based on its related database object.

    May also process an item while also taking into account its parent item
    e.g. a Column (child item) within a Model (parent item)
    """
    overwrite: bool = Field(
        description=(
            "Whether to overwrite existing properties with properties from the database. "
            "When false, keeps the values already present in the properties if present."
        ),
        default=True,
        examples=[True, False],
    )

    @abstractmethod
    def run(self, source: S, target: T) -> bool:
        """
        Run this generator on the given source, merging properties from the target item.

        :param source: The item to modify.
        :param target: The item containing the target properties to set onto the source.
        :return: Boolean for if the source was modified.
        """
        raise NotImplementedError


class PropertiesGenerator[I: ItemT, G: PropertyGenerator](ContractPart, metaclass=ABCMeta):
    """
    A part of a contract meant to generate properties for a resource based on its related database object.

    May also process an item while also taking into account its parent item
    e.g. a Column (child item) within a Model (parent item)
    """
    exclude: Annotated[Sequence[str], BeforeValidator(to_tuple)] = Field(
        description="The fields to exclude from the generated properties.",
        default=(),
    )

    @classmethod
    def _name(cls) -> str:
        """The name of this generator in snake_case."""
        class_name = cls.__name__.replace("Properties", "").replace("Generator", "")
        return re.sub(r"([a-z])([A-Z])", r"\1_\2", class_name).lower()

    @property
    def generators(self) -> list[PropertyGenerator]:
        """
        Get all the single property generators set on this generator,
        excluding those that are configured to be excluded.
        """
        return [
            attr for name, attr in vars(self).items()
            if name not in self.exclude and isinstance(attr, PropertyGenerator)
        ]

    @abstractmethod
    def merge(self, item: I, context: ContractContext) -> bool:
        """
        Merge the properties of the given item with the properties of its remote database object.

        :param item: The item to modify properties for.
        :param context: The contract context to use.
        :return: Whether the properties of the item were modified.
        """
        raise NotImplementedError


class ChildPropertiesGenerator[I: ItemT, P: PropertiesT, G: PropertyGenerator](
    PropertiesGenerator[I, G], metaclass=ABCMeta
):
    @abstractmethod
    def merge(self, item: I, context: ContractContext, parent: P = None) -> bool:
        """
        Merge the properties of the given item with the properties of its remote database object.

        :param item: The item to modify properties for.
        :param context: The contract context to use.
        :param parent: The parent item that the given child `item` belongs to if available.
        :return: Whether the properties of the item were modified.
        """
        raise NotImplementedError


class ParentPropertiesGenerator[I: PropertiesT, G: PropertyGenerator](PropertiesGenerator[I, G], metaclass=ABCMeta):
    _properties_defaults: dict[str, Any] = {"version": 2}

    filename: str = Field(
        description="The name to give to new properties files generated by this generator.",
        default="_config.yml",
        examples=["properties.yml", "config.yml"],
    )
    depth: int | None = Field(
        description=(
            "The depth at which to place newly generated files within the resource folder. "
            "e.g. 0 would place the file in the root of the resource folder, "
            "1 would place it in a subfolder of the root of the resource folder, etc. "
            "By default, the file is placed in the same folder in which the resource is stored."
        ),
        default=None,
        ge=0,
        examples=[0, 1, 2, 3],
    )

    def update(self, item: I, context: ContractContext) -> dict[str, Any]:
        """
        Update/generate the properties of the given item as a properties file.

        If the item already has a properties path configured, properties will be updated in-place.
        Otherwise, a new properties file will be generated based on the configuration of this generator.
        This new properties file will be added to the context's properties store.
        The path for this new properties file will be added to the item's attributes.

        :param item: The item to update properties for.
        :param context: The contract context to use.
        :return: The updated properties.
        """
        try:
            path = context.properties.get_path(item, to_absolute=True)
        except FileNotFoundError:
            path = None

        if path is not None and path.is_file():
            return self._update_existing_properties(item, properties=context.properties[path])

        flags = get_flags()
        project_dir = Path(getattr(flags, "PROJECT_DIR", None) or "")

        path = self.generate_properties_path(item)
        if isinstance(item, ParsedResource):
            item.patch_path = f"{context.manifest.metadata.project_name}://{path.relative_to(project_dir)}"
        elif isinstance(item, BaseResource):
            item.original_file_path = str(path.relative_to(project_dir))

        if (existing_properties_file := context.properties.get(path)) is not None:
            # file exists but item's properties are not in it
            # update properties file with the item's generated properties
            return self._update_existing_properties(item, properties=existing_properties_file)

        properties = self._generate_properties(item)
        context.properties[path] = properties
        return properties

    @abstractmethod
    def _update_existing_properties(self, item: I, properties: dict[str, Any]) -> dict[str, Any]:
        raise NotImplementedError

    @staticmethod
    @abstractmethod
    def _generate_properties(item: Any) -> dict[str, Any]:
        """Generate a properties mapping for the given item."""
        raise NotImplementedError

    def generate_properties_path(self, item: I) -> Path:
        """Generate a new properties path to use for the given item."""
        flags = get_flags()
        project_dir = Path(getattr(flags, "PROJECT_DIR", None) or "")

        item_path = item.original_file_path or item.path
        if self.depth is None:
            path = project_dir.joinpath(item_path).with_name(self.filename)
        else:
            parts = Path(item_path).parts
            path = project_dir.joinpath(*parts[:self.depth + 1]).joinpath(self.filename)

        if path.suffix.casefold() not in {".yml", ".yaml"}:
            path = path.with_suffix(".yml")

        return path



from pathlib import Path
from random import choice

import pytest
from _pytest.fixtures import FixtureRequest
from dbt.contracts.graph.nodes import ModelNode
from faker import Faker

from dbt_contracts.contracts.conditions.properties import NameCondition, PathCondition, TagCondition, MetaCondition, \
    IsMaterializedCondition


@pytest.mark.parametrize("item", ["model", "column", "argument"])
def test_name_validation(item: str, faker: Faker, request: FixtureRequest):
    item = request.getfixturevalue(item)

    assert NameCondition().run(item)
    assert NameCondition(include=faker.words() + [item.name]).run(item)
    assert not NameCondition(exclude=faker.words() + [item.name]).run(item)


def test_path_validation(model: ModelNode, faker: Faker):
    paths = [faker.file_path() for _ in range(5)]

    assert PathCondition().run(model)
    assert PathCondition(include=paths + [model.path]).run(model)
    assert not PathCondition(exclude=paths + [model.patch_path.split("://")[1]]).run(model)
    assert not PathCondition(
        include=paths + [model.path], exclude=paths + [model.patch_path.split("://")[1]]
    ).run(model)


def test_path_condition_escapes_backslashes_in_paths():
    paths = ["path\\to\\folder1", "path\\to\\file1"]
    expected = [path.replace("\\", "\\\\") for path in paths]
    assert PathCondition.escape_backslashes_in_windows_paths(paths) == expected


def test_path_condition_unifies_chunked_path_values():
    # no remapping for these values
    assert PathCondition.unify_chunked_path_values("path") == ("path",)
    assert PathCondition.unify_chunked_path_values(["path1", "path2"]) == ("path1", "path2")
    assert PathCondition.unify_chunked_path_values([["path1"], ["path2"]]) == ("path1", "path2")

    paths = [["path", "to", "folder1"]]
    assert PathCondition.unify_chunked_path_values(paths) == (str(Path(*paths[0])),)

    paths = [["path", "to", "folder1"], ["path", "to", "folder2"], "path/to/folder3"]
    expected = tuple(str(Path(*parts)) if not isinstance(parts, str) else parts for parts in paths)
    assert PathCondition.unify_chunked_path_values(paths) == expected

    condition = PathCondition(include=paths, exclude=paths)
    assert condition.include == expected
    assert condition.exclude == expected


@pytest.mark.parametrize("item", ["model", "column"])
def test_tag_validation(item: str, faker: Faker, request: FixtureRequest):
    item = request.getfixturevalue(item)

    assert TagCondition().run(item)
    assert not TagCondition(tags=[word for word in faker.words() if word not in item.tags]).run(item)
    assert TagCondition(tags=faker.words() + [choice(item.tags)]).run(item)


@pytest.mark.parametrize("item", ["model", "column"])
def test_meta_validation(item: str, faker: Faker, request: FixtureRequest):
    item = request.getfixturevalue(item)
    meta = {key: faker.words() for key in item.meta}

    assert MetaCondition().run(item)
    assert not MetaCondition(meta=meta).run(item)

    key, value = choice(list(item.meta.items()))
    meta[key].append(value)
    assert not MetaCondition(meta=meta).run(item)

    for key, value in item.meta.items():
        meta[key].append(value)
    assert MetaCondition(meta=meta).run(item)


def test_is_materialized_validation(model: ModelNode, faker: Faker):
    model.config.materialized = "view"
    assert IsMaterializedCondition().run(model)

    model.config.materialized = "ephemeral"
    assert not IsMaterializedCondition().run(model)



from unittest import mock

from dbt.artifacts.resources.v1.macro import MacroArgument
from dbt.contracts.graph.nodes import Macro

from dbt_contracts.contracts import ContractContext
from dbt_contracts.contracts.terms.macro import HasType


def test_argument_has_type(argument: MacroArgument, macro: Macro, context: ContractContext):
    argument.type = "int"
    assert HasType().run(argument, parent=macro, context=context)

    argument.type = None
    with mock.patch.object(ContractContext, "add_result") as mock_add_result:
        assert not HasType().run(argument, parent=macro, context=context)
        mock_add_result.assert_called_once()



"""
Generic types to use for all contracts.
"""
from dbt.artifacts.resources.base import BaseResource
from dbt.artifacts.resources.v1.components import ColumnInfo, ParsedResource
from dbt.artifacts.resources.v1.macro import MacroArgument
from dbt.contracts.graph.nodes import SourceDefinition, Macro

type ChildT = ColumnInfo | MacroArgument
type ParentT = BaseResource | None
type ItemT = ColumnInfo | MacroArgument | BaseResource | None

type PropertiesT = ParsedResource | SourceDefinition | Macro
type DescriptionT = ParsedResource | ColumnInfo | SourceDefinition | Macro | MacroArgument
type TagT = ParsedResource | ColumnInfo
type MetaT = ParsedResource | ColumnInfo
type NodeT = ParsedResource | SourceDefinition



from abc import ABCMeta, abstractmethod
from collections.abc import Mapping
from pathlib import Path
from typing import Any, Self, ClassVar

from dbt.artifacts.resources import BaseResource
from dbt.artifacts.resources.v1.components import ColumnInfo
from dbt.artifacts.resources.v1.macro import MacroArgument
from dbt.contracts.graph.nodes import ModelNode, SourceDefinition, Macro

from dbt_contracts._core import BaseModelConfig
from dbt_contracts.properties import PropertiesIO
from dbt_contracts.types import ItemT, ParentT


class Result[I: ItemT, P: ParentT](BaseModelConfig, metaclass=ABCMeta):
    """Store information of the result from a contract execution."""
    name: str
    path: Path | None
    result_type: str
    result_level: str
    result_name: str
    message: str
    # properties attributes
    properties_path: Path | None = None
    properties_start_line: int | None = None
    properties_start_col: int | None = None
    properties_end_line: int | None = None
    properties_end_col: int | None = None
    # parent specific attributes
    parent_id: str | None = None
    parent_name: str | None = None
    parent_type: str | None = None
    index: int | None = None

    resource_type: ClassVar[type[ItemT]]

    @property
    def has_parent(self) -> bool:
        """Was this result built using a parent item."""
        return self.parent_id is not None or self.parent_name is not None or self.parent_type is not None

    @classmethod
    def from_resource(
            cls, item: I, parent: P = None, properties: PropertiesIO = None, **kwargs
    ) -> Self:
        """
        Create a new :py:class:`Result` from a given resource.

        :param item: The resource to log a result for.
        :param parent: The parent item that the given child `item` belongs to if available.
        :param properties: The properties IO handler to use when loading properties files.
        :return: The :py:class:`Result` instance.
        """
        props = {}
        props_path = None
        props_item = parent if parent is not None else item
        if properties is not None and (item_properties := properties.get(props_item)) is not None:
            props = cls._extract_properties_for_item(item_properties, item=item, parent=parent) or {}
            props_path = properties.get_path(props_item, to_absolute=False)

        if parent is not None:
            kwargs |= dict(
                parent_id=parent.unique_id,
                parent_name=parent.name,
                parent_type=parent.resource_type.name.title(),
            )

        path = None
        if isinstance(path_item := parent if parent is not None else item, BaseResource):
            path = Path(path_item.original_file_path)

        # noinspection PyUnresolvedReferences
        field_names: set[str] = set(cls.model_fields.keys())

        return cls(
            name=item.name,
            path=path,
            result_type=cls._get_result_type(item=item, parent=parent),
            properties_path=props_path,
            properties_start_line=props.get("__start_line__"),
            properties_start_col=props.get("__start_col__"),
            properties_end_line=props.get("__end_line__"),
            properties_end_col=props.get("__end_col__"),
            **{key: val for key, val in kwargs.items() if key in field_names},
        )

    @staticmethod
    def _get_result_type(item: I, parent: P = None) -> str:
        result_type = item.resource_type.name.title()
        if parent:
            result_type = f"{parent.resource_type.name.title()} {result_type}"
        return result_type

    @classmethod
    @abstractmethod
    def _extract_properties_for_item(
            cls, properties: Mapping[str, Any], item: ItemT, parent: ParentT = None
    ) -> Mapping[str, Any] | None:
        raise NotImplementedError

    def as_github_annotation(self) -> Mapping[str, str]:
        """
        Format this result to a GitHub annotation. Raises an exception if the result does not
        have all the required parameters set to build a valid GitHub annotation.
        """
        if not self.can_format_to_github_annotation:
            raise Exception("Cannot format this result to a GitHub annotation.")
        return self._as_github_annotation()

    @property
    def can_format_to_github_annotation(self) -> bool:
        """Can this result be formatted as a valid GitHub annotation."""
        required_keys = {"path", "start_line", "end_line", "annotation_level", "message"}
        annotation = self._as_github_annotation()
        return all(annotation.get(key) is not None for key in required_keys)

    def _as_github_annotation(self) -> Mapping[str, str | int | list[str] | dict[str, str]]:
        """
        See annotations spec in the `output` param 'Update a check run':
        https://docs.github.com/en/rest/checks/runs?apiVersion=2022-11-28#update-a-check-run
        """
        return {
            "path": str(self.properties_path or self.path),
            "start_line": self.properties_start_line,
            "start_column": self.properties_start_col,
            "end_line": self.properties_end_line,
            "end_column": self.properties_end_col,
            "annotation_level": self.result_level,
            "title": self.result_name.replace("_", " ").title(),
            "message": self.message,
            "raw_details": {
                "result_type": self.result_type,
                "name": self.name,
            },
        }


class ModelResult(Result[ModelNode, None]):
    resource_type = ModelNode

    @classmethod
    def _extract_properties_for_item(
            cls, properties: Mapping[str, Any], item: ModelNode, parent: None = None
    ) -> Mapping[str, Any] | None:
        models = (model for model in properties.get("models", ()) if model.get("name", "") == item.name)
        return next(models, None)


class SourceResult(Result[SourceDefinition, None]):
    resource_type = SourceDefinition

    @classmethod
    def _extract_properties_for_item(
            cls, properties: Mapping[str, Any], item: SourceDefinition, parent: None = None
    ) -> Mapping[str, Any] | None:
        sources = (
            table
            for source in properties.get("sources", ()) if source.get("name", "") == item.source_name
            for table in source.get("tables", ()) if table.get("name", "") == item.name
        )
        return next(sources, None)


class ColumnResult[P: ParentT](Result[ColumnInfo, P]):
    resource_type = ColumnInfo

    @classmethod
    def _extract_properties_for_item(
            cls, properties: Mapping[str, Any], item: ColumnInfo, parent: P = None
    ) -> Mapping[str, Any] | None:
        result_processor = RESULT_PROCESSOR_MAP.get(type(parent))
        if result_processor is None:
            return

        # noinspection PyProtectedMember
        parent_properties = result_processor._extract_properties_for_item(properties=properties, item=parent)
        if parent_properties is None:
            return

        columns = (column for column in parent_properties.get("columns", ()) if column.get("name", "") == item.name)
        return next(columns, None)

    @classmethod
    def _get_result_type(cls, item: ColumnInfo, parent: P = None) -> str:
        result_type = "Column"
        if parent is not None:
            result_type = f"{parent.resource_type.name.title()} {result_type}"
        return result_type

    @classmethod
    def from_resource(
            cls, item: ColumnInfo, parent: P = None, properties: PropertiesIO = None, **kwargs
    ) -> Self:
        try:
            index = list(parent.columns.keys()).index(item.name) if parent is not None else None
        except ValueError:
            index = None

        return super().from_resource(item=item, parent=parent, index=index, **kwargs)


class MacroResult(Result[Macro, None]):
    resource_type = Macro

    @classmethod
    def _extract_properties_for_item(
            cls, properties: Mapping[str, Any], item: Macro, parent: None = None
    ) -> Mapping[str, Any] | None:
        macros = (macro for macro in properties.get("macros", ()) if macro.get("name", "") == item.name)
        return next(macros, None)


class MacroArgumentResult(Result[MacroArgument, Macro]):
    resource_type = MacroArgument

    @classmethod
    def _extract_properties_for_item(
            cls, properties: Mapping[str, Any], item: MacroArgument, parent: Macro = None
    ) -> Mapping[str, Any] | None:
        # noinspection PyProtectedMember
        macro = MacroResult._extract_properties_for_item(properties=properties, item=parent)
        if macro is None:
            return

        arguments = (argument for argument in macro.get("arguments", ()) if argument.get("name", "") == item.name)
        return next(arguments, None)

    @classmethod
    def _get_result_type(cls, item: MacroArgument, parent: Macro = None) -> str:
        return "Macro Argument"

    @classmethod
    def from_resource(
            cls, item: MacroArgument, parent: Macro = None, properties: PropertiesIO = None, **kwargs
    ) -> Self:
        index = parent.arguments.index(item) if parent is not None else None
        return super().from_resource(item=item, parent=parent, index=index, **kwargs)


RESULT_PROCESSORS: tuple[type[Result], ...] = (
    ModelResult, SourceResult, MacroResult, ColumnResult, MacroArgumentResult
)
RESULT_PROCESSOR_MAP: Mapping[type[ItemT], type[Result]] = {cls.resource_type: cls for cls in RESULT_PROCESSORS}



from abc import ABCMeta
from collections.abc import Mapping, Sequence
from random import choice, sample
from typing import Literal, Any, Annotated, get_args

from dbt.artifacts.resources.v1.components import ColumnInfo
from dbt_common.contracts.metadata import ColumnMetadata, CatalogTable
from pydantic import Field, BeforeValidator

from dbt_contracts.contracts._core import ContractContext
from dbt_contracts.contracts.generators._core import ParentPropertiesGenerator, CORE_FIELDS, PropertyGenerator
from dbt_contracts.contracts.generators.properties import SetDescription
from dbt_contracts.contracts.utils import get_matching_catalog_table, merge_maps, to_tuple
from dbt_contracts.types import NodeT

NODE_FIELDS = Literal[CORE_FIELDS, "columns"]


class NodePropertyGenerator[S: NodeT](PropertyGenerator[S, CatalogTable], metaclass=ABCMeta):
    pass


class SetNodeDescription[S: NodeT](SetDescription[S, CatalogTable], NodePropertyGenerator[S]):
    def run(self, source: S, target: CatalogTable) -> bool:
        return self._set_description(source, description=target.metadata.comment)


class SetNodeColumns[S: NodeT](NodePropertyGenerator[S]):
    add: bool = Field(
        description=(
            "Add columns to the properties file which are in the database object but missing from the properties."
        ),
        default=True,
        examples=[True, False],
    )
    remove: bool = Field(
        description=(
            "Remove columns from the properties file which are in the properties file but not in the database object."
        ),
        default=True,
        examples=[True, False],
    )
    order: bool = Field(
        description="Reorder columns in the properties file to match the order found in the database object.",
        default=True,
        examples=[True, False],
    )

    @classmethod
    def _name(cls) -> str:
        return "columns"

    @staticmethod
    def _set_column(source: S, column: ColumnMetadata) -> bool:
        if any(col.name == column.name for col in source.columns.values()):
            return False

        source.columns[column.name] = ColumnInfo(name=column.name)
        return True

    @staticmethod
    def _drop_column(source: S, column: ColumnInfo, columns: Mapping[str, ColumnMetadata]) -> bool:
        if any(col.name == column.name for col in columns.values()):
            return False

        source.columns.pop(column.name)
        return True

    @staticmethod
    def _order_columns(source: S, columns: Mapping[str, ColumnMetadata]):
        index_map = {col.name: col.index for col in columns.values()}
        columns_in_order = dict(
            sorted(source.columns.items(), key=lambda col: index_map.get(col[1].name, len(index_map)))
        )
        if list(columns_in_order) == list(source.columns):
            return False

        source.columns.clear()
        source.columns.update(columns_in_order)
        return True

    def run(self, source: S, target: CatalogTable) -> bool:
        if not target.columns:
            return False

        columns = target.columns

        added = False
        if self.add:
            added = any([self._set_column(source, column=column) for column in columns.values()])

        removed = False
        if self.remove:
            removed = any([
                self._drop_column(source, column=column, columns=columns) for column in source.columns.copy().values()
            ])

        ordered = False
        if self.order:
            ordered = self._order_columns(source, target.columns)

        return added or removed or ordered


EXCLUDE_TYPES = Literal["description", "columns"]


class NodePropertiesGenerator(ParentPropertiesGenerator[NodeT, NodePropertyGenerator], metaclass=ABCMeta):
    exclude: Annotated[Sequence[EXCLUDE_TYPES], BeforeValidator(to_tuple)] = Field(
        description="The fields to exclude from the generated properties.",
        default=(),
        examples=[choice(get_args(EXCLUDE_TYPES)), sample(get_args(EXCLUDE_TYPES), k=2)]
    )
    description: SetNodeDescription = Field(
        description="Configuration for setting the description",
        default=SetNodeDescription(),
    )
    columns: SetNodeColumns = Field(
        description="Configuration for setting the columns",
        default=SetNodeColumns(),
    )

    def merge(self, item: NodeT, context: ContractContext) -> bool:
        if (table := get_matching_catalog_table(item, catalog=context.catalog)) is None:
            return False

        return any([generator.run(item, table) for generator in self.generators])

    @classmethod
    def _merge_columns(cls, item: NodeT, table: dict[str, Any]) -> None:
        if "columns" not in table:
            table["columns"] = []

        for column in table["columns"].copy():
            if column["name"] not in item.columns:
                table["columns"].remove(column)

        for index, column_info in enumerate(item.columns.values()):
            column = cls._generate_column_properties(column_info)
            index_in_props, column_in_props = next(
                ((i, col) for i, col in enumerate(table["columns"]) if col["name"] == column_info.name),
                (None, None)
            )

            if column_in_props is not None:
                merge_maps(column_in_props, column, overwrite=True, extend=False)
            else:
                table["columns"].insert(index, column)

            if index_in_props is not None and index_in_props != index:
                table["columns"].pop(index_in_props)
                table["columns"].insert(index, column_in_props)

    @staticmethod
    def _generate_column_properties(column: ColumnInfo) -> dict[str, Any]:
        column = {
            "name": column.name,
            "description": column.description,
            "data_type": column.data_type,
        }

        return {key: val for key, val in column.items() if val}



import json
from argparse import Namespace
from pathlib import Path
from unittest import mock

import pytest
from dbt.artifacts.schemas.catalog import CatalogArtifact
from dbt.cli.main import dbtRunner, dbtRunnerResult
from dbt.constants import MANIFEST_FILE_NAME
from dbt.contracts.graph.manifest import Manifest
from dbt.task.docs.generate import CATALOG_FILENAME

from dbt_contracts.dbt_cli import add_default_args, load_artifact, DEFAULT_GLOBAL_ARGS, get_result, clean_paths, \
    install_dependencies, get_manifest, get_catalog


@pytest.fixture
def config(tmp_path: Path) -> Namespace:
    """Fixture for dbt config."""
    config = Namespace(
        project_root="project_root",
        args=Namespace(profiles_dir="profiles_dir"),
        profile_name="profile_name",
        target_name="target_name",
        project_target_path=str(tmp_path.joinpath("target")),
    )

    with mock.patch("dbt_contracts.dbt_cli.get_config", return_value=config):
        yield config


def test_add_default_args(config: Namespace):
    assert add_default_args("arg1", "arg2") == ["arg1", "arg2"]

    with mock.patch("dbt_contracts.dbt_cli.get_config", return_value=config):
        assert add_default_args("arg1", "arg2", "--profile", "test", config=config) == [
            "arg1",
            "arg2",
            "--profile", "test",
            "--project-dir", "project_root",
            "--profiles-dir", "profiles_dir",
            "--target", "target_name"
        ]


def test_load_artifact(config: Namespace):
    filename = "manifest.json"
    expected = {"key": "value"}

    assert load_artifact(filename, config) is None

    path = Path(config.project_target_path, filename)
    path.parent.mkdir(parents=True)

    assert load_artifact(filename, config) is None

    with path.open("w") as file:
        json.dump(expected, file)

    assert load_artifact(filename, config) == expected


def test_get_result():
    result = dbtRunnerResult(success=True)
    with mock.patch.object(dbtRunner, "invoke", return_value=result) as mock_invoke:
        assert get_result("arg1", "arg2") == result
        mock_invoke.assert_called_once_with(["arg1", "arg2", *DEFAULT_GLOBAL_ARGS])

        result.success = False
        result.exception = Exception
        with pytest.raises(result.exception):
            get_result("arg1", "arg2")


def test_clean_paths(config: Namespace):
    runner = dbtRunner()

    with mock.patch("dbt_contracts.dbt_cli.get_result") as mock_get_result:
        clean_paths("arg1", "arg2", runner=runner, config=config)
        args = add_default_args("arg1", "arg2", config=config)
        mock_get_result.assert_called_once_with("clean", "--no-clean-project-files-only", *args, runner=runner)


def test_install_dependencies(config: Namespace):
    runner = dbtRunner()

    with mock.patch("dbt_contracts.dbt_cli.get_result") as mock_get_result:
        install_dependencies("arg1", "arg2", runner=runner, config=config)
        args = add_default_args("arg1", "arg2", config=config)
        mock_get_result.assert_called_once_with("deps", *args, runner=runner)


def test_get_manifest_with_existing_artifact(config: Namespace, manifest: Manifest):
    manifest = manifest.to_dict()
    with (
        mock.patch("dbt_contracts.dbt_cli.load_artifact", return_value=manifest) as mock_load_artifact,
        mock.patch("dbt_contracts.dbt_cli.get_result") as mock_get_result,
    ):
        assert get_manifest(config=config).to_dict() == manifest
        mock_load_artifact.assert_called_once_with(MANIFEST_FILE_NAME, config=config)
        mock_get_result.assert_not_called()


def test_get_manifest_with_no_existing_artifact(config: Namespace):
    runner = dbtRunner()

    with mock.patch("dbt_contracts.dbt_cli.get_result") as mock_get_result:
        get_manifest("arg1", "arg2", runner=runner, config=config)
        args = add_default_args("arg1", "arg2", config=config)
        mock_get_result.assert_called_once_with("parse", *args, runner=runner)


def test_get_manifest_with_existing_artifact_with_refresh(config: Namespace, catalog: CatalogArtifact):
    catalog = catalog.to_dict()
    with (
        mock.patch("dbt_contracts.dbt_cli.load_artifact", return_value=catalog) as mock_load_artifact,
        mock.patch("dbt_contracts.dbt_cli.get_result") as mock_get_result,
    ):
        get_manifest(config=config, refresh=True)
        mock_load_artifact.assert_not_called()
        mock_get_result.assert_called_once()


def test_get_catalog_with_existing_artifact(config: Namespace, catalog: CatalogArtifact):
    catalog = catalog.to_dict()
    with (
        mock.patch("dbt_contracts.dbt_cli.load_artifact", return_value=catalog) as mock_load_artifact,
        mock.patch("dbt_contracts.dbt_cli.get_result") as mock_get_result,
    ):
        assert get_catalog(config=config).to_dict() == catalog
        mock_load_artifact.assert_called_once_with(CATALOG_FILENAME, config=config)
        mock_get_result.assert_not_called()


def test_get_catalog_with_existing_artifact_with_refresh(config: Namespace, catalog: CatalogArtifact):
    catalog = catalog.to_dict()
    with (
        mock.patch("dbt_contracts.dbt_cli.load_artifact", return_value=catalog) as mock_load_artifact,
        mock.patch("dbt_contracts.dbt_cli.get_result") as mock_get_result,
    ):
        get_catalog(config=config, refresh=True)
        mock_load_artifact.assert_not_called()
        mock_get_result.assert_called_once()


def test_get_catalog_with_no_existing_artifact(config: Namespace):
    runner = dbtRunner()

    with mock.patch("dbt_contracts.dbt_cli.get_result") as mock_get_result:
        get_catalog("arg1", "arg2", runner=runner, config=config)
        args = add_default_args("arg1", "arg2", config=config)
        mock_get_result.assert_called_once_with("docs", "generate", *args, runner=runner)



from abc import ABCMeta

from dbt_common.dataclass_schema import dbtClassMixin
from pydantic import Field

from dbt_contracts.contracts.generators._core import PropertyGenerator
from dbt_contracts.types import DescriptionT


class SetDescription[S: DescriptionT, T: dbtClassMixin](PropertyGenerator[S, T], metaclass=ABCMeta):
    terminator: str | None = Field(
        description=(
            "Only take the description up to this terminating string. "
            "e.g. if you only want to take the first line of a multi-line description, set this to '\\n'"
        ),
        default=None,
        examples=["\\n", "__END__", "."],
    )

    @classmethod
    def _name(cls) -> str:
        return "description"

    def _set_description(self, item: S, description: str | None) -> bool:
        if not description:
            return False
        if item.description and not self.overwrite:
            return False

        if self.terminator:
            description = description.split(self.terminator)[0]
        if item.description == description:
            return False

        item.description = description
        return True



from random import sample

import pytest
from dbt.contracts.graph.nodes import ModelNode

from dbt_contracts.contracts import ContractContext
from dbt_contracts.contracts.generators.model import ModelPropertiesGenerator
from tests.contracts.generators.test_node import NodePropertiesGeneratorTester


class TestModelPropertiesGenerator(NodePropertiesGeneratorTester[ModelNode]):
    @pytest.fixture
    def generator(self) -> ModelPropertiesGenerator:
        return ModelPropertiesGenerator()

    @pytest.fixture
    def item(self, model: ModelNode) -> ModelNode:
        return model

    def test_generate_table_properties(self, generator: ModelPropertiesGenerator, item: ModelNode):
        table = generator._generate_table_properties(item)
        assert all(val for val in table.values())

    def test_generate_properties(self, generator: ModelPropertiesGenerator, item: ModelNode):
        properties = generator._generate_properties(item)
        assert item.resource_type.pluralize() in properties

        columns = list(map(generator._generate_column_properties, item.columns.values()))
        table = generator._generate_table_properties(item) | {"columns": columns}

        assert table in properties[item.resource_type.pluralize()]
        for key, val in generator._properties_defaults.items():
            assert properties[key] == val

    def test_update_existing_properties_with_empty_properties(
            self, generator: ModelPropertiesGenerator, item: ModelNode, context: ContractContext
    ):
        key = item.resource_type.pluralize()
        properties = {}
        expected_table = generator._generate_properties(item)[key][0]

        generator._update_existing_properties(item, properties=properties)
        assert len(properties[key]) == 1
        assert expected_table in properties[key]

    def test_update_existing_properties_with_new_table(
            self,
            generator: ModelPropertiesGenerator,
            item: ModelNode,
            models: list[ModelNode],
            context: ContractContext
    ):
        key = item.resource_type.pluralize()
        models = sample([model for model in models if model.name != item.name], k=5)
        properties = {key: [models[key][0] for models in map(generator._generate_properties, models)]}
        original_count = len(properties[key])

        expected_columns = list(map(generator._generate_column_properties, item.columns.values()))
        expected_table = generator._generate_table_properties(item) | {"columns": expected_columns}

        generator._update_existing_properties(item, properties=properties)
        assert len(properties[key]) == original_count + 1
        assert expected_table in properties[key]

    def test_update_existing_properties_with_existing_table(
            self,
            generator: ModelPropertiesGenerator,
            item: ModelNode,
            models: list[ModelNode],
            context: ContractContext
    ):
        key = item.resource_type.pluralize()
        models = sample([model for model in models if model.name != item.name], k=5)
        table = generator._generate_table_properties(item)
        properties = {key: [models[key][0] for models in map(generator._generate_properties, models)] + [table]}
        original_count = len(properties[key])

        # should update the description in the properties
        item.description = "a brand new description"

        generator._update_existing_properties(item, properties=properties)
        assert len(properties[key]) == original_count

        actual_tables = [prop for prop in properties[key] if prop["name"] == item.name]
        assert len(actual_tables) == 1
        assert actual_tables[0]["description"] == item.description



from random import sample
from unittest import mock

import pytest
from dbt.artifacts.resources.types import TimePeriod
from dbt.artifacts.resources.v1.components import FreshnessThreshold, Time
from dbt.contracts.graph.nodes import SourceDefinition

from dbt_contracts.contracts import ContractContext
from dbt_contracts.contracts.terms.source import HasLoader, HasFreshness, HasDownstreamDependencies


def test_has_loader(source: SourceDefinition, context: ContractContext):
    source.loader = "i am a loader"
    assert HasLoader().run(source, context=context)

    source.loader = ""
    with mock.patch.object(ContractContext, "add_result") as mock_add_result:
        assert not HasLoader().run(source, context=context)
        mock_add_result.assert_called_once()


@pytest.fixture
def fresh_source(source: SourceDefinition) -> SourceDefinition:
    """Fixture for a source with freshness"""
    source.loaded_at_field = "i am a loaded at field"
    source.freshness = FreshnessThreshold(
        warn_after=Time(count=2, period=TimePeriod.hour),
        error_after=Time(count=5, period=TimePeriod.hour),
    )
    return source


def test_has_freshness(fresh_source: SourceDefinition, context: ContractContext):
    assert HasFreshness().run(fresh_source, context=context)


def test_has_no_freshness_invalid_loaded_at_field(fresh_source: SourceDefinition, context: ContractContext):
    fresh_source.loaded_at_field = None

    with mock.patch.object(ContractContext, "add_result") as mock_add_result:
        assert not HasFreshness().run(fresh_source, context=context)
        mock_add_result.assert_called_once()


def test_has_no_freshness_invalid_freshness(fresh_source: SourceDefinition, context: ContractContext):
    fresh_source.freshness.warn_after.count = None
    fresh_source.freshness.error_after.period = None

    with mock.patch.object(ContractContext, "add_result") as mock_add_result:
        assert not HasFreshness().run(fresh_source, context=context)
        mock_add_result.assert_called_once()


def test_has_downstream_dependencies(source: SourceDefinition, context: ContractContext):
    downstream_deps = sample(
        [node for node in context.manifest.nodes.values() if node.unique_id.startswith("model")], k=3
    )
    assert downstream_deps
    for dep in downstream_deps:
        dep.depends_on_nodes.append(source.unique_id)

    assert HasDownstreamDependencies().run(source, context=context)

    with mock.patch.object(ContractContext, "add_result") as mock_add_result:
        assert not HasDownstreamDependencies(min_count=5).run(source, context=context)
        mock_add_result.assert_called_once()



from dbt.contracts.graph.nodes import SourceDefinition
from faker import Faker

from dbt_contracts.contracts.conditions.source import IsEnabledCondition


def test_is_enabled_validation(source: SourceDefinition, faker: Faker):
    source.config.enabled = True
    assert IsEnabledCondition().run(source)

    source.config.enabled = False
    assert not IsEnabledCondition().run(source)



import itertools
import re
import textwrap
from collections.abc import Sequence, Callable, Iterable, Collection
from typing import Self, Any, Annotated, Literal

from colorama import Fore
from pydantic import Field, model_validator, BeforeValidator

from dbt_contracts._core import BaseModelConfig
from dbt_contracts.contracts.result import Result
from dbt_contracts.contracts.utils import to_tuple
from dbt_contracts.formatters import ResultsFormatter


def _get_print_length(value: str) -> int:
    pattern = re.compile(r'\x1B(?:[@-Z\\-_]|\[[0-?]*[ -/]*[@-~])')
    return len(pattern.sub("", value))


def _align_and_pad_print_length(value: str, alignment: Literal["<", "^", ">"], width: int) -> str:
    width += len(value) - _get_print_length(value)
    return f"{value:{alignment}{width}}"


class TableCellBuilder[T: Result](BaseModelConfig):
    key: str | Callable[[T], str] = Field(
        description="The key to use to get the value from the :py:class:`.Result` "
                    "or a callable object which returns a formatted string for the value",
    )
    prefix: str | None = Field(
        description=(
            "The prefix to prepend to the value. "
            "This prefix will always be left aligned regardless of the alignment setting."
        ),
        default=None
    )
    alignment: Literal["<", "^", ">"] = Field(
        description="The alignment of the value in the cell.",
        default="<"
    )
    colour: str | None = Field(
        description="The colour to apply to the value.",
        default=None
    )
    min_width: int | None = Field(
        description="The minimum width of the cell. Any value shorter than this will pad the cell.",
        default=None,
        ge=1,
    )
    max_width: int | None = Field(
        description="The maximum width of the cell. Any value larger than this will be truncated with '…'.",
        default=None,
        ge=1,
    )
    wrap: bool = Field(
        description="Whether to wrap the value in the cell.",
        default=False
    )

    @model_validator(mode="after")
    def validate_wrap(self) -> Self:
        """Validate the required parameters are set when using wrap=True."""
        if self.wrap and not self.max_width:
            raise Exception("Cannot wrap a cell without a max width.")
        return self

    @property
    def prefix_coloured(self) -> str:
        """Get the prefix with colour applied if set."""
        if not self.prefix:
            return ""
        if not self.colour:
            return self.prefix
        return f"{self.colour.replace("m", ";1m")}{self.prefix}{Fore.RESET.replace("m", ";0m")}"

    def _get_value(self, result: T) -> str:
        value = self.key(result) if callable(self.key) else getattr(result, self.key, "")
        if value is None:
            return ""
        return str(value)

    def _apply_prefix(self, value: str) -> str:
        if not self.prefix or not _get_print_length(value):
            return value
        return f"{self.prefix_coloured}{value}"

    def _truncate_value(self, value: str) -> str:
        width = self.max_width
        if not width or _get_print_length(value) <= width:
            return value

        if self.prefix_coloured:
            width -= len(self.prefix)
        return value[:width - 1] + "…"

    def _apply_padding_and_alignment(self, value: str, width: int | None, has_prefix: bool = True) -> str:
        if not self.alignment or not width:
            return value

        if has_prefix and self.prefix:
            width -= len(self.prefix)

        return _align_and_pad_print_length(value, self.alignment, width)

    def _apply_colour(self, value: str) -> str:
        if not self.colour:
            return value
        return f"{self.colour}{value}{Fore.RESET}"

    def _apply_wrap(self, value: str) -> list[str]:
        if not self.wrap or not self.max_width:
            return [value]

        lines = textwrap.wrap(
            (self.prefix if self.prefix else "") + value,
            width=self.max_width,
            break_long_words=False,
            break_on_hyphens=False
        )
        if self.prefix:
            lines[0] = lines[0][len(self.prefix):]

        return lines

    def build(self, result: T, min_width: int = None) -> str:
        """
        Build a formatted cell for the given `result`.

        :param result: The result to build the cell for.
        :param min_width: Ignore settings for min width and force the cell to be at least this given width instead.
        :return: The formatted cell.
        """
        min_width = min_width if min_width is not None else self.min_width

        value = self._get_value(result)
        if self.wrap and self.max_width:
            lines = self._apply_wrap(value)
            lines = list(map(self._apply_colour, lines))
            lines = [
                self._apply_padding_and_alignment(line, width=min_width, has_prefix=i == 0)
                for i, line in enumerate(lines)
            ]
            lines[0] = self._apply_prefix(lines[0])
            value = "\n".join(lines)
        else:
            value = self._truncate_value(value)
            value = self._apply_colour(value)
            value = self._apply_padding_and_alignment(value, width=min_width)
            value = self._apply_prefix(value)

        return value


class TableRowBuilder[T: Result](BaseModelConfig):
    cells: Sequence[TableCellBuilder[T]] | Sequence[Sequence[TableCellBuilder[T] | None]] = Field(
        description="The cell builders to use to build the row.",
    )
    separator: str = Field(
        description="The separator to use between the cells.",
        default="|"
    )
    colour: str | None = Field(
        description="The colour to apply to the separator.",
        default=None
    )

    @model_validator(mode="after")
    def remap_and_validate_cells(self) -> Self:
        """Remap the cells if they are not in the matrix format and validate the cells."""
        if all(isinstance(cell, TableCellBuilder) for cell in self.cells):
            self.cells = [self.cells]

        if not len(set(map(len, self.cells))) == 1:
            raise Exception("All cell rows must be of equal length")
        if not all(isinstance(cell, TableCellBuilder) for cell in self.cells[0]):
            raise Exception("All cells in the 1st row must be filled")

        return self

    @property
    def separator_coloured(self) -> str:
        """Get the separator with colour applied if set."""
        if not self.colour:
            return self.separator
        return f"{self.colour.replace("m", ";1m")}{self.separator}{Fore.RESET.replace("m", ";0m")}"

    def _get_lines(self, result: Result, min_widths: Sequence[int | None]) -> list[str]:
        values: list[list[str]] = []  # col -> row list

        for cells in self.cells:
            line: list[str] = [
                cell.build(result, min_width=min_width) if cell is not None else ""
                for cell, min_width in zip(cells, min_widths, strict=True)
            ]
            values.append(line)

            lines = [value.splitlines() for value in map("\n".join, zip(*values, strict=True))]
            min_widths = self.get_widths_from_lines(list(zip(*lines)))

        return list(map("\n".join, zip(*values, strict=True)))

    @staticmethod
    def _get_max_rows(values: Iterable[str]) -> int:
        return max(len(value.splitlines()) for value in values)

    @classmethod
    def _to_matrix(cls, values: Sequence[str]) -> list[tuple[str]]:
        """
        Expand the lines in the given values, ensuring that a valid NxM matrix of values is returned.

        :param values: The values of the cells in the row.
        :return: The matrix of cells for each row.
        """
        row_count = cls._get_max_rows(values)

        columns = []
        for value in values:
            rows = value.splitlines()
            rows.extend([""] * (row_count - len(rows)))
            columns.append(rows)

        return list(zip(*columns, strict=True))

    @staticmethod
    def _remove_empty_lines[T: Iterable[str]](lines: Iterable[T]) -> list[T]:
        """
        Remove empty lines from the given `values`.

        :param lines: The values to remove empty lines from.
        """
        return [line for line in lines if any(bool(cell.strip()) for cell in line)]

    @staticmethod
    def get_widths_from_lines(lines: Iterable[Iterable[str]]) -> list[int]:
        """Get the maximum width for each column from the given `lines`."""
        return [max(map(_get_print_length, column)) for column in zip(*lines, strict=True)]

    def extend_line_widths(
            self, rows: Iterable[Collection[Iterable[str]]], min_widths: Sequence[int | None]
    ) -> list[list[list[str]]]:
        """
        Adjust the given `rows` to the given `min_widths` aligning according to the matching cell config.
        Each row must consist of a sequence of lines that match the dimensions of the configured cells of this builder.

        :param rows: The collection of rows to adjust.
        :param min_widths: The widths to extend the lines to.
        :return: The extended lines.
        """
        # noinspection PyTypeChecker
        if len(min_widths) != len(self.cells[0]):
            # noinspection PyTypeChecker
            raise Exception(
                f"Given widths do not equal the number of cells configured for this row builder: "
                f"cells={len(self.cells[0])}, widths={len(min_widths)}"
            )

        result: list[list[list[str]]] = []
        for lines in rows:
            cells_padded = list(self.cells) + list(itertools.repeat(self.cells[-1], len(lines) - len(self.cells)))
            result_lines: list[list[str]] = []

            for line, cells in zip(lines, cells_padded, strict=True):
                result_line: list[str] = []

                for value, cell, width in zip(line, cells, min_widths, strict=True):
                    if not value and width is not None:
                        value = " " * width
                    elif width is not None:
                        alignment = cell.alignment if cell is not None else self.cells[0][0].alignment
                        value = _align_and_pad_print_length(value, alignment, width)

                    result_line.append(value)
                result_lines.append(result_line)
            result.append(result_lines)

        return result

    def build_lines(self, result: T, min_widths: Sequence[int | None] = ()) -> list[list[str]]:
        """
        Build a formatted row or set of rows for the given `result` and returns them as their individual lines.

        :param result: The result to build the row for.
        :param min_widths: When provided, ignore settings for min width for each cell and force the cells to be
            at least these given widths instead.
            Must contain the same number of elements as there are cells configured for this builder.
        :return: The formatted row as a set of lines.
        """
        if not min_widths:
            # noinspection PyTypeChecker
            min_widths = [None] * len(self.cells[0])
        # noinspection PyTypeChecker
        if len(min_widths) != len(self.cells[0]):
            raise Exception(
                f"Given widths do not equal the number of cells configured for this row builder: "
                f"cells={len(self.cells)}, widths={len(min_widths)}"
            )

        lines = self._get_lines(result, min_widths=min_widths)
        lines = self._to_matrix(lines)
        min_widths = self.get_widths_from_lines(lines)
        lines = self.extend_line_widths([lines], min_widths=min_widths)[0]
        lines = self._remove_empty_lines(lines)
        return lines

    def join(self, rows: Iterable[Iterable[Iterable[str]]]) -> str:
        """Join the given rows into a single string."""
        return "\n".join(map(f" {self.separator_coloured} ".join, itertools.chain.from_iterable(rows)))

    def build(self, result: T, min_widths: Sequence[int] = ()) -> str:
        """
        Build a formatted row or set of rows for the given `result`.

        :param result: The result to build the row for.
        :param min_widths: When provided, ignore settings for min width for each cell and force the cells to be
            at least these given widths instead.
            Must contain the same number of elements as there are cells configured for this builder.
        :return: The formatted row.
        """
        lines = self.build_lines(result, min_widths=min_widths)
        return self.join([lines])


class TableFormatter[T: Result](ResultsFormatter[T]):
    builder: TableRowBuilder[T] = Field(
        description="The builder for the rows of the table",
    )
    consistent_widths: bool = Field(
        description="Whether to ensure all rows have the same width for each column.",
        default=False,
    )

    @property
    def widths(self) -> list[int]:
        """Get the max widths of each column in the currently stored result"""
        flattened_results = itertools.chain.from_iterable(self._results)
        return self.builder.get_widths_from_lines(flattened_results)

    def __init__(self, /, **data: Any):
        super().__init__(**data)

        self._lines: list[str] = []
        self._results: list[list[list[str]]] = []

    def add_header(self, header: str) -> None:
        """Add a header to the table."""
        if len(self._lines) >= 2 and self._lines[1] == "":
            self._lines.pop(0)
            self._lines.pop(0)

        self._lines.insert(0, "")
        self._lines.insert(0, header)

    def add_results(self, results: Iterable[T]) -> None:
        for result in results:
            widths = self.widths if self.consistent_widths else None
            row = self.builder.build_lines(result, min_widths=widths)
            self._results.append(row)

        if self.consistent_widths:
            self._results = self.builder.extend_line_widths(self._results, min_widths=self.widths)

    def build(self) -> str:
        output = "\n".join(self._lines) + self.builder.join(self._results)

        self._lines.clear()
        self._results.clear()

        return output


class GroupedTableFormatter[T: Result](ResultsFormatter[T]):
    formatter: TableFormatter[T] = Field(
        description="The formatter to use for each table.",
    )
    group_key: str | Callable[[T], str] = Field(
        description="The key to use to get the group value for each table from a :py:class:`.Result` "
                    "or a callable object which returns a formatted string for the value",
    )
    header_key: str | Callable[[T], str] | None = Field(
        description="The key to use to get the header value for each table from a :py:class:`.Result` "
                    "or a callable object which returns a formatted string for the value",
        default=None,
    )
    sort_key: Annotated[Sequence[str | Callable[[T], Any]], BeforeValidator(to_tuple)] = Field(
        description="The key to use to get the sort value for each :py:class:`.Result` in a table "
                    "or a callable object which returns a formatted string for the value",
        default=None,
    )

    def __init__(self, /, **data: Any):
        super().__init__(**data)

        self._tables: dict[str, str] = {}

    @staticmethod
    def _get_value(result: T, getter: str | Callable[[T], Any]) -> Any:
        if callable(getter):
            return getter(result)
        return getattr(result, getter, "") or ""

    @classmethod
    def _get_values(cls, result: T, getters: Iterable[str | Callable[[T], Any]]) -> tuple[Any, ...]:
        return tuple(cls._get_value(result, getter=getter) for getter in getters)

    def add_results(self, results: Iterable[T]) -> None:
        results = sorted(results, key=lambda r: self._get_value(result=r, getter=self.group_key))
        groups = itertools.groupby(results, key=lambda r: self._get_value(result=r, getter=self.group_key))

        for group_key, group in groups:
            if self.sort_key:
                group = sorted(group, key=lambda r: self._get_values(result=r, getters=self.sort_key))
            else:
                group = list(group)
            header = self._get_value(result=group[0], getter=self.header_key) if self.header_key else group_key

            self.formatter.add_header(header)
            self.formatter.add_results(group)

            self._tables[group_key] = self.formatter.build()

    def build(self) -> str:
        output = "\n\n".join(self._tables.values())
        self._tables.clear()
        return output



"""
Fills in the variable fields of the README template and generates README.md file.
"""
import re
from random import sample

import docstring_parser
import yaml

from dbt_contracts import PROGRAM_OWNER_USER, PROGRAM_NAME, DOCUMENTATION_URL
from dbt_contracts.contracts import CONTRACT_CLASSES, Contract, ParentContract
import docs.contracts as docs

from dbt_contracts.contracts.generators import PropertiesGenerator
from dbt_contracts.runner import ContractsRunner

SRC_FILENAME = "README.template.md"
TRG_FILENAME = SRC_FILENAME.replace(".template", "")


def format_contract_title(contract: type[Contract], parent_key: str = "") -> str:
    """Format the title for a contract"""
    key = f"{parent_key.rstrip('s')}_{contract.__config_key__}"
    return key.replace("_", " ").title().strip()


def format_contracts_example() -> str:
    contracts = {}

    for cls in sample(CONTRACT_CLASSES, k=2):
        example = docs.ReferencePageBuilder.generate_example_dict_for_contract(cls)
        if issubclass(cls, ParentContract):
            child_cls = cls.__child_contract__
            child_example = docs.ReferencePageBuilder.generate_example_dict_for_contract(child_cls)
            example[child_cls.__config_key__] = [child_example]

        contracts[cls.__config_key__] = [example]

    example = {"contracts": contracts}

    return f"   ```yaml\n{yaml.dump(example, sort_keys=False).strip()}\n```".replace("\n", "\n   ")


def format_contract_reference(contract: type[Contract], parent_key: str = "") -> list[str]:
    """Format the readme template for the contracts reference"""
    lines = []

    key = contract.__config_key__
    title = format_contract_title(contract, parent_key)
    lines.extend((f"### {title}", ""))

    # noinspection PyTypeChecker,PyProtectedMember
    contract_parts_map = {
        "Filters": contract.__supported_conditions__,
        "Terms": contract.__supported_terms__,
        "Generator": contract.__supported_generator__,
    }

    for header, parts in contract_parts_map.items():
        if parts is None:
            continue
        elif issubclass(parts, PropertiesGenerator):
            url = f"{DOCUMENTATION_URL}/{'/'.join(docs.URL_PATH)}/{key}.html#{header.lower()}"
            line = (
                f"You may also [configure a generator]({url}) to automatically and dynamically "
                f"generate properties files for these {key} from database objects."
            )
            lines.append(line)
            continue

        lines.extend((f"#### {header}", ""))
        for part in parts:
            # noinspection PyProtectedMember
            name = part._name()
            url = f"{DOCUMENTATION_URL}/{'/'.join(docs.URL_PATH)}/{key}.html#{name.replace('_', '-')}"
            doc = docstring_parser.parse(part.__doc__).short_description.strip().format(kind=title.lower())
            doc = re.sub(r"\s*\n\s+", " ", doc)

            line = f"- [`{name}`]({url}): {doc}"
            lines.append(line)

        lines.append("")

    lines.append("")

    if issubclass(contract, ParentContract):
        lines.extend(format_contract_reference(contract.__child_contract__, key))

    return lines


def format_contracts_reference() -> str:
    """Format the readme template for the contracts reference"""
    lines = []
    for contract in CONTRACT_CLASSES:
        lines.extend(format_contract_reference(contract))

    return "\n".join(lines)


def format_contracts_reference_toc_entry(contract: type[Contract], parent_key: str = "") -> str:
    """Format the readme template for a contracts reference table of contents entry"""
    title = format_contract_title(contract, parent_key)
    return f"  * [{title}](#{title.replace(" ", "-").lower()})"


def format_contracts_reference_toc() -> str:
    """Format the readme template for the contracts reference table of contents"""
    lines = []

    for contract in CONTRACT_CLASSES:
        lines.append(format_contracts_reference_toc_entry(contract))
        if issubclass(contract, ParentContract):
            lines.append(format_contracts_reference_toc_entry(contract.__child_contract__, contract.__config_key__))

    return "\n".join(lines)


def format_readme():
    """Format the readme template and save the formatted readme"""
    format_map_standard = {
        "program_name": PROGRAM_NAME,
        "program_name_lower": PROGRAM_NAME.lower(),
        "program_owner_user": PROGRAM_OWNER_USER,
        "documentation_url": DOCUMENTATION_URL,
        "default_contracts_filename": ContractsRunner.default_config_file_name,
        "contracts_example": format_contracts_example(),
        "contracts_reference": format_contracts_reference().strip(),
        "contracts_reference_toc": format_contracts_reference_toc().rstrip(),
    }
    format_map_code = {
    }
    format_map_code = {k: "`" + "` `".join(v) + "`" for k, v in format_map_code.items()}
    format_map = format_map_standard | format_map_code

    with open(SRC_FILENAME, 'r') as file:
        template = file.read()

    formatted = template.format_map(format_map)
    with open(TRG_FILENAME, 'w') as file:
        file.write(formatted)


if __name__ == "__main__":
    format_readme()
    print(f"Formatted {TRG_FILENAME} file using template: {SRC_FILENAME}")



from random import sample
from unittest import mock

import pytest
from _pytest.fixtures import FixtureRequest
from dbt.contracts.graph.nodes import SourceDefinition
from faker import Faker

from dbt_contracts.contracts import ContractContext
from dbt_contracts.contracts.terms.properties import HasProperties, HasDescription, HasRequiredTags, HasAllowedTags, \
    HasRequiredMetaKeys, HasAllowedMetaKeys, HasAllowedMetaValues
from dbt_contracts.types import PropertiesT, DescriptionT, TagT, MetaT


@pytest.mark.parametrize("item", ["model", "source", "macro"])
def test_has_properties(item: str, context: ContractContext, faker: Faker, request: FixtureRequest):
    item: PropertiesT = request.getfixturevalue(item)

    item.patch_path = faker.file_path()
    assert HasProperties().run(item, context=context)

    item.patch_path = None
    if isinstance(item, SourceDefinition):  # always returns true for sources
        assert HasProperties().run(item, context=context)
    else:
        with mock.patch.object(ContractContext, "add_result") as mock_add_result:
            assert not HasProperties().run(item, context=context)
            mock_add_result.assert_called_once()


@pytest.mark.parametrize("item", ["model", "source", "column", "macro", "argument"])
def test_has_description(item: str, context: ContractContext, faker: Faker, request: FixtureRequest):
    item: DescriptionT = request.getfixturevalue(item)

    item.description = faker.sentence()
    assert HasDescription().run(item, context=context)

    item.description = ""
    with mock.patch.object(ContractContext, "add_result") as mock_add_result:
        assert not HasDescription().run(item, context=context)
        mock_add_result.assert_called_once()


@pytest.mark.parametrize("item", ["model", "column"])
def test_has_required_tags(item: str, context: ContractContext, faker: Faker, request: FixtureRequest):
    item: TagT = request.getfixturevalue(item)

    item.tags = faker.words(10)
    assert HasRequiredTags(tags=sample(item.tags, k=5)).run(item, context=context)
    with mock.patch.object(ContractContext, "add_result") as mock_add_result:
        assert not HasRequiredTags(tags=faker.words(10) + sample(item.tags, k=5)).run(item, context=context)
        mock_add_result.assert_called_once()

    item.tags.clear()
    assert HasRequiredTags().run(item, context=context)
    with mock.patch.object(ContractContext, "add_result") as mock_add_result:
        assert not HasRequiredTags(tags=faker.words(10)).run(item, context=context)
        mock_add_result.assert_called_once()


@pytest.mark.parametrize("item", ["model", "column"])
def test_has_allowed_tags(item: str, context: ContractContext, faker: Faker, request: FixtureRequest):
    item: TagT = request.getfixturevalue(item)

    item.tags = faker.words(10)
    with mock.patch.object(ContractContext, "add_result") as mock_add_result:
        assert not HasAllowedTags(tags=sample(item.tags, k=5)).run(item, context=context)
        mock_add_result.assert_called_once()

        assert not HasAllowedTags(tags=faker.words(10)).run(item, context=context)
        assert len(mock_add_result.mock_calls) == 2

    assert HasAllowedTags(tags=faker.words(10) + item.tags).run(item, context=context)

    item.tags.clear()
    assert HasAllowedTags().run(item, context=context)
    assert HasAllowedTags(tags=faker.words(10)).run(item, context=context)


@pytest.mark.parametrize("item", ["model", "column"])
def test_has_required_meta_keys(item: str, context: ContractContext, faker: Faker, request: FixtureRequest):
    item: MetaT = request.getfixturevalue(item)

    item.meta = {key: faker.word() for key in faker.words(10)}
    assert HasRequiredMetaKeys(keys=sample(list(item.meta), k=5)).run(item, context=context)
    with mock.patch.object(ContractContext, "add_result") as mock_add_result:
        assert not HasRequiredMetaKeys(keys=faker.words(10) + sample(list(item.meta), k=5)).run(item, context=context)
        mock_add_result.assert_called_once()

    item.meta.clear()
    assert HasRequiredMetaKeys().run(item, context=context)

    with mock.patch.object(ContractContext, "add_result") as mock_add_result:
        assert not HasRequiredMetaKeys(keys=faker.words(10)).run(item, context=context)
        mock_add_result.assert_called_once()


@pytest.mark.parametrize("item", ["model", "column"])
def test_has_allowed_meta_keys(item: str, context: ContractContext, faker: Faker, request: FixtureRequest):
    item: MetaT = request.getfixturevalue(item)

    item.meta = {key: faker.word() for key in faker.words(10)}
    with mock.patch.object(ContractContext, "add_result") as mock_add_result:
        assert not HasAllowedMetaKeys(keys=sample(list(item.meta), k=5)).run(item, context=context)
        mock_add_result.assert_called_once()

        assert not HasAllowedMetaKeys(keys=faker.words(10)).run(item, context=context)
        assert len(mock_add_result.mock_calls) == 2

    assert HasAllowedMetaKeys(keys=faker.words(10) + list(item.meta)).run(item, context=context)

    item.meta.clear()
    assert HasAllowedMetaKeys().run(item, context=context)
    assert HasAllowedMetaKeys(keys=faker.words(10)).run(item, context=context)


@pytest.mark.parametrize("item", ["model", "column"])
def test_has_allowed_meta_values(item: str, context: ContractContext, faker: Faker, request: FixtureRequest):
    item: MetaT = request.getfixturevalue(item)

    item.meta = {key: faker.word() for key in faker.words(10)}
    allowed_meta_values = {key: faker.words(5) for key, val in sample(list(item.meta.items()), k=5)}
    with mock.patch.object(ContractContext, "add_result") as mock_add_result:
        assert not HasAllowedMetaValues(meta=allowed_meta_values).run(item, context=context)
        mock_add_result.assert_called_once()

    allowed_meta_values = {key: faker.words(5) + [val] for key, val in sample(list(item.meta.items()), k=5)}
    assert HasAllowedMetaValues(meta=allowed_meta_values).run(item, context=context)

    allowed_meta_values = {key: faker.words(3) for key in faker.words(10) if key not in item.meta}
    assert HasAllowedMetaValues(meta=allowed_meta_values).run(item, context=context)

    item.meta.clear()
    assert HasAllowedMetaValues().run(item, context=context)
    assert HasAllowedMetaValues(meta=allowed_meta_values).run(item, context=context)



import json
from argparse import Namespace
from pathlib import Path
from random import choice, sample
from unittest import mock

import pytest
import yaml
from faker import Faker

from dbt_contracts.contracts import ModelContract, SourceContract, ColumnContract
from dbt_contracts.contracts.conditions import properties as c_properties
from dbt_contracts.contracts.result import Result, ModelResult
from dbt_contracts.contracts.terms import properties as t_properties, source as t_source, column as t_column
from dbt_contracts.properties import PropertiesIO
# noinspection PyProtectedMember
from dbt_contracts.runner import _get_default_table_header, ContractsRunner


class TestContractsRunner:
    @pytest.fixture
    def results(self, tmp_path: Path) -> list[Result]:
        """A fixture for the results."""
        return [
            ModelResult(
                name="this is the 1st result",
                path=tmp_path,
                result_type="failure",
                result_level="error",
                result_name="result",
                message="This is an error message.",
            ),
            ModelResult(
                name="this is the 5th result",
                path=tmp_path,
                result_type="warning",
                result_level="warning",
                result_name="result",
                message="This is a warning message.",
            ),
            ModelResult(
                name="this is the 3rd result",
                path=tmp_path,
                result_type="another very great success",
                result_level="info",
                result_name="result",
                message="This is a success message.",
            ),
            ModelResult(
                name="this is the 4th result",
                path=tmp_path,
                result_type="failure",
                result_level="error",
                result_name="result",
                message="This is an error message.",
            ),
            ModelResult(
                name="this is the 2nd result",
                path=tmp_path,
                result_type="a very great success",
                result_level="info",
                result_name="result",
                message="This is a success message.",
            ),
        ]

    @pytest.fixture
    def runner(self, tmp_path: Path) -> ContractsRunner:
        """Fixture for the contracts runner to test."""
        contracts = []
        conditions = [
            c_properties.NameCondition(include=["model1", "model2"]),
            c_properties.TagCondition(tags=["include"])
        ]
        terms = [
            t_properties.HasRequiredTags(tags="required_tag"),
        ]
        contracts.append(ModelContract(conditions=conditions, terms=terms))

        conditions = [
            c_properties.NameCondition(include=["source1", "source2"]),
            c_properties.TagCondition(tags=["include"])
        ]
        terms = [
            t_properties.HasRequiredTags(tags="required_tag"),
            t_source.HasLoader(),
        ]
        contracts.append(SourceContract(conditions=conditions, terms=terms))

        conditions = [
            c_properties.TagCondition(tags=["valid"])
        ]
        terms = [
            t_column.HasDataType(),
        ]
        contracts.append(ColumnContract(parent=contracts[0], conditions=conditions, terms=terms))

        for contract in contracts:
            assert not contract.needs_manifest
            assert not contract.needs_catalog

        config = Namespace(
            project_root=str(tmp_path),
            project_name="name",
        )

        # noinspection PyTypeChecker
        return ContractsRunner(contracts=contracts, config=config)

    def test_get_default_table_header(self, results: list[Result]):
        """Test the `get_default_table_header` function."""
        result = results[0]
        assert not result.properties_path

        output = _get_default_table_header(result)
        assert result.result_type in output
        assert str(result.path) in output
        assert str(result.properties_path) not in output

        result.properties_path = Path("properties_path.yml")
        output = _get_default_table_header(result)
        assert result.result_type in output
        assert str(result.path) in output
        assert str(result.properties_path) in output

    def test_cached_properties(self, runner: ContractsRunner):
        config = runner.config
        assert runner.config is config

        dbt = runner.dbt
        assert runner.dbt is dbt

        with mock.patch("dbt_contracts.runner.dbt_cli.get_manifest", return_value="manifest"):
            manifest = runner.manifest
            assert runner.manifest is manifest
            assert runner.dbt.manifest is manifest

        with mock.patch("dbt_contracts.runner.dbt_cli.get_catalog", return_value="catalog"):
            catalog = runner.catalog
            assert runner.catalog is catalog

    def test_set_paths(self, runner: ContractsRunner):
        def _get_absolute_path(p: str | Path) -> Path:
            path = Path(p)
            if path == paths[-1]:
                return path
            return Path(runner.config.project_root, path)

        paths = [
            str(Path("path", "to", "model.sql")),
            Path("path", "to", "another", "model.sql"),
            Path("path", "to", "dir"),
            Path("path", "not", "in", "project"),
        ]

        with mock.patch("dbt_contracts.runner.get_absolute_project_path", new=_get_absolute_path):
            runner.paths = paths

        assert runner.paths.include == tuple(map(str, paths[:3]))

        with mock.patch("dbt_contracts.runner.get_absolute_project_path", new=_get_absolute_path):
            runner.paths = paths[0]
        assert runner.paths.include == tuple(map(str, paths[:1]))

        runner.paths = []
        assert runner.paths is None

    def test_resolve_config_path(self, tmp_path: Path):
        path = tmp_path.joinpath("path", "to", "dir")
        path.mkdir(parents=True)

        expected = path.joinpath(ContractsRunner.default_config_file_name)
        expected.touch()

        assert ContractsRunner._resolve_config_path(path) == expected

    # noinspection PyTestUnpassedFixture
    def test_from_objects(self):
        args = Namespace(config="path/to/config.yml")
        config = Namespace(args=args)

        with (
            mock.patch.object(ContractsRunner, "from_file") as mock_file,
            mock.patch("dbt_contracts.runner.dbt_cli.get_config", return_value="config") as mock_config
        ):
            # noinspection PyTypeChecker
            ContractsRunner.from_config(config)
            mock_file.assert_called_with(config.args.config, config=config)

            ContractsRunner.from_args(args)
            mock_file.assert_called_with(args.config, config=mock_config.return_value)

    def test_from_file(self, tmp_path: Path):
        path = tmp_path.joinpath(ContractsRunner.default_config_file_name)
        path.with_suffix("")

        with (
            mock.patch.object(ContractsRunner, "from_yaml") as mock_yaml,
            mock.patch.object(ContractsRunner, "from_json") as mock_json,
        ):
            # uses yml extension by default when no extension given
            # noinspection PyTypeChecker
            ContractsRunner.from_file(path, config=None)
            mock_yaml.assert_called_once()
            mock_json.assert_not_called()

            # noinspection PyTypeChecker
            ContractsRunner.from_file(path.with_suffix(choice((".yml", ".yaml"))), config=None)
            assert len(mock_yaml.mock_calls) == 2
            mock_json.assert_not_called()

            # noinspection PyTypeChecker
            ContractsRunner.from_file(path.with_suffix(".json"), config=None)
            assert len(mock_yaml.mock_calls) == 2
            mock_json.assert_called_once()

        with pytest.raises(Exception):
            # noinspection PyTypeChecker
            ContractsRunner.from_file(path.with_suffix(".mp3"), config=None)

    def test_from_yaml(self, tmp_path: Path):
        path = tmp_path.joinpath("config.yml")
        config = {"models": [{"filters": [], "terms": []}]}
        with path.open("w") as f:
            yaml.dump(config, f)

        with mock.patch.object(ContractsRunner, "from_dict") as mock_dict:
            # noinspection PyTypeChecker
            ContractsRunner.from_yaml(path, config=None)
            mock_dict.assert_called_once_with(config, config=None)

    def test_from_json(self, tmp_path: Path):
        path = tmp_path.joinpath("config.json")
        config = {"models": [{"filters": [], "terms": []}]}
        with path.open("w") as f:
            json.dump(config, f)

        with mock.patch.object(ContractsRunner, "from_dict") as mock_dict:
            # noinspection PyTypeChecker
            ContractsRunner.from_json(path, config=None)
            mock_dict.assert_called_once_with(config, config=None)

    def test_from_dict(self):
        config = {
            "contracts": {
                "models": [
                    {
                        "filters": sample([cls._name() for cls in ModelContract.__supported_conditions__], k=3),
                        "terms": sample([cls._name() for cls in ModelContract.__supported_terms__], k=3),
                    },
                    {
                        "filters": sample([cls._name() for cls in ModelContract.__supported_conditions__], k=3),
                        "terms": sample([cls._name() for cls in ModelContract.__supported_terms__], k=3),
                    }
                ],
                "sources": [
                    {
                        "filters": sample([cls._name() for cls in SourceContract.__supported_conditions__], k=3),
                        "terms": sample([cls._name() for cls in SourceContract.__supported_terms__], k=3),
                    }
                ]
            }
        }

        with mock.patch.object(ContractsRunner, "_create_contracts_from_config") as mock_create:
            # noinspection PyTypeChecker
            ContractsRunner.from_dict(config, config=None)

            assert len(mock_create.mock_calls) == 3 * 2  # *2 for iter calls
            mock_create.assert_any_call("models", config=config["contracts"]["models"][0])
            mock_create.assert_any_call("models", config=config["contracts"]["models"][1])
            mock_create.assert_any_call("sources", config=config["contracts"]["sources"][0])

    def test_create_contracts_from_config(self):
        models_config = {
            "filters": sample([cls._name() for cls in ModelContract.__supported_conditions__], k=3),
            "terms": sample([cls._name() for cls in ModelContract.__supported_terms__], k=3),
        }
        sources_config = {
            "filters": sample([cls._name() for cls in ModelContract.__supported_conditions__], k=3),
            "terms": sample([cls._name() for cls in ModelContract.__supported_terms__], k=3),
        }
        columns_config = {
            "filters": sample([cls._name() for cls in ColumnContract.__supported_conditions__], k=3),
            "terms": sample([cls._name() for cls in ColumnContract.__supported_terms__], k=3),
        }

        with (
            mock.patch.object(ModelContract, "from_dict", return_value=ModelContract()) as mock_model,
            mock.patch.object(SourceContract, "from_dict", return_value=SourceContract()) as mock_source,
            mock.patch.object(ColumnContract, "from_dict", return_value=ColumnContract()) as mock_column,
        ):
            ContractsRunner._create_contracts_from_config("model", models_config)
            mock_model.assert_called_once()
            mock_source.assert_not_called()
            mock_column.assert_not_called()

            ContractsRunner._create_contracts_from_config("source", sources_config)
            mock_model.assert_called_once()
            mock_source.assert_called_once()
            mock_column.assert_not_called()

            sources_config[ColumnContract.__config_key__] = columns_config
            ContractsRunner._create_contracts_from_config("source", sources_config)
            mock_column.assert_called_once()

    def test_set_artifacts_on_contracts_skips_artifacts(self, runner: ContractsRunner):
        # manifest is set by default as it will always be available
        runner.__dict__["manifest"] = "manifest"

        runner._set_artifacts_on_contracts(runner._contracts)

        # cached properties were not set
        assert "catalog" not in runner.__dict__

        for contract in runner._contracts:
            assert not contract.needs_catalog
            assert contract.catalog is None

    def test_set_artifacts_on_contracts_sets_artifacts(self, runner: ContractsRunner):
        # set cached properties manually
        runner.__dict__["manifest"] = "manifest"
        runner.__dict__["catalog"] = "catalog"

        contract = choice(list(runner._contracts))
        contract.terms.append(
            choice([term() for term in contract.__supported_terms__ if term.needs_manifest])
        )
        contract.terms.append(
            choice([term() for term in contract.__supported_terms__ if term.needs_catalog])
        )
        assert contract.needs_manifest
        assert contract.needs_catalog

        runner._set_artifacts_on_contracts(runner._contracts)
        for c in runner._contracts:
            assert c.manifest == "manifest"
            assert c.catalog == "catalog"

    def test_set_artifacts_on_contracts_sets_paths(self, runner: ContractsRunner):
        runner.__dict__["manifest"] = "manifest"
        runner._paths = c_properties.PathCondition(include="path/to/model.sql")

        runner._set_artifacts_on_contracts(runner._contracts)
        for contract in runner._contracts:
            if c_properties.PathCondition in contract.__supported_conditions__:
                assert runner.paths in contract.conditions
            else:
                assert runner.paths not in contract.conditions

        # only adds the condition once
        runner._set_artifacts_on_contracts(runner._contracts)
        for contract in runner._contracts:
            if c_properties.PathCondition in contract.__supported_conditions__:
                assert contract.conditions.count(runner.paths) == 1
            else:
                assert contract.conditions.count(runner.paths) == 0

    def test_get_contract_by_key(self, runner: ContractsRunner):
        contract = choice(list(runner._contracts))
        assert runner._get_contract_by_key(contract.config_key) == contract

        with pytest.raises(Exception):
            runner._get_contract_by_key("unknown key")

    def test_validate_runs_all_contracts(self, runner: ContractsRunner):
        with (
            mock.patch.object(ModelContract, "validate") as mock_model,
            mock.patch.object(SourceContract, "validate") as mock_source,
            mock.patch.object(ColumnContract, "validate") as mock_column,
            mock.patch.object(ContractsRunner, "_set_artifacts_on_contracts") as mock_set_artifacts,
        ):
            runner.validate()

            mock_set_artifacts.assert_called_once()
            mock_model.assert_called_once()
            mock_source.assert_called_once()
            mock_column.assert_called_once()

    def test_validate_runs_selected_contract(self, runner: ContractsRunner):
        runner.__dict__["manifest"] = "manifest"

        with (
            mock.patch.object(ModelContract, "validate") as mock_model,
            mock.patch.object(SourceContract, "validate") as mock_source,
            mock.patch.object(ColumnContract, "validate") as mock_column,
        ):
            key = ModelContract.child_config_key
            terms = ["term_1", "term_2"]
            runner.validate(key, terms=terms)

            mock_model.assert_not_called()
            mock_source.assert_not_called()
            mock_column.assert_called_once_with(terms=terms)

    def test_validate_logs_results(self, runner: ContractsRunner, results: list[Result]):
        runner.__dict__["manifest"] = "manifest"
        assert all(not contract.context.results for contract in runner._contracts)

        with (
            mock.patch.object(ModelContract, "validate"),
            mock.patch.object(SourceContract, "validate"),
            mock.patch.object(ColumnContract, "validate"),
            mock.patch.object(ContractsRunner, "log_results") as mock_log_results,
        ):
            assert not runner.validate()
            mock_log_results.assert_not_called()
            assert all(not contract.context.results for contract in runner._contracts)

            contract = choice(list(runner._contracts))
            contract.context.results.extend(results.copy())

            assert runner.validate() == results
            mock_log_results.assert_called_once()
            mock_log_results.assert_any_call(results)
            assert all(not contract.context.results for contract in runner._contracts)

    def test_generate_runs_all_contracts(self, runner: ContractsRunner, tmp_path: Path):
        with (
            mock.patch.object(ModelContract, "generate", return_value={tmp_path: 1}) as mock_model,
            mock.patch.object(SourceContract, "generate", return_value={tmp_path: 2}) as mock_source,
            mock.patch.object(ColumnContract, "generate", return_value={tmp_path: 3}) as mock_column,
            mock.patch.object(ContractsRunner, "_set_artifacts_on_contracts") as mock_set_artifacts,
            mock.patch.object(ContractsRunner, "_log_generated_paths") as mock_log_paths,
            mock.patch.object(PropertiesIO, "save") as mock_save,
            mock.patch("dbt_contracts.runner.dbt_cli.get_manifest", return_value="manifest") as mock_manifest,
        ):
            results = runner.generate()
            assert results == {tmp_path: 6}

            mock_set_artifacts.assert_any_call(runner._contracts, force=True)
            mock_model.assert_called_once()
            mock_source.assert_called_once()
            mock_column.assert_called_once()

            mock_log_paths.assert_called_once_with({})  # nothing was saved so results log is empty
            mock_save.assert_called_with(results)
            mock_manifest.assert_any_call(runner=runner.dbt, config=runner.config, logger=runner.logger, refresh=True)
            mock_set_artifacts.assert_called_with(runner._contracts)

    def test_generate_runs_selected_contract(self, runner: ContractsRunner, tmp_path: Path):
        with (
            mock.patch.object(ModelContract, "generate", return_value={tmp_path: 1}) as mock_model,
            mock.patch.object(SourceContract, "generate", return_value={tmp_path: 3}) as mock_source,
            mock.patch.object(ColumnContract, "generate", return_value={tmp_path: 8}) as mock_column,
            mock.patch.object(ContractsRunner, "_set_artifacts_on_contracts"),
            mock.patch.object(ContractsRunner, "_log_generated_paths") as mock_log_paths,
            mock.patch.object(PropertiesIO, "save", return_value=[tmp_path]) as mock_save,
            mock.patch("dbt_contracts.runner.dbt_cli.get_manifest", return_value="manifest") as mock_manifest,
        ):
            key = ModelContract.child_config_key
            results = runner.generate(key)
            assert results == {tmp_path: 8}

            mock_model.assert_not_called()
            mock_source.assert_not_called()
            mock_column.assert_called_once()

            mock_log_paths.assert_called_once_with(results)
            mock_save.assert_called_with(results)
            mock_manifest.assert_any_call(runner=runner.dbt, config=runner.config, logger=runner.logger, refresh=True)

    def test_build_results(self, runner: ContractsRunner, results: list[Result]):
        assert runner._build_results([]) == ""
        assert runner._build_results(results) != ""

    def test_write_results(self, runner: ContractsRunner, results: list[Result], tmp_path: Path):
        output_type = choice(list(runner.output_writers_map))
        assert runner.write_results([], path=tmp_path, output_type=output_type) is None

        output_path = runner.write_results(results, path=tmp_path, output_type="json")
        assert output_path == tmp_path.joinpath(runner.default_output_file_name).with_suffix(".json")
        assert output_path.is_file()

        # uses given filename when not dir
        output_path = tmp_path.joinpath("file.json")
        assert runner.write_results(results, path=output_path, output_type="json") == output_path

    def test_write_results_as_text(self, runner: ContractsRunner, results: list[Result], tmp_path: Path):
        output_path = runner.write_results(results, path=tmp_path, output_type="text")
        assert output_path == tmp_path.joinpath(runner.default_output_file_name).with_suffix(".txt")

        with output_path.open("r") as f:
            assert f.read() == runner._build_results(results)

    def test_write_results_as_json(self, runner: ContractsRunner, results: list[Result], tmp_path: Path):
        output_path = runner.write_results(results, path=tmp_path, output_type="json")
        assert output_path == tmp_path.joinpath(runner.default_output_file_name).with_suffix(".json")

        expected = json.dumps([result.model_dump_json() for result in results], indent=2)
        with output_path.open("r") as f:
            assert f.read() == expected

    def test_write_results_as_jsonl(self, runner: ContractsRunner, results: list[Result], tmp_path: Path):
        output_path = runner.write_results(results, path=tmp_path, output_type="jsonl")
        assert output_path == tmp_path.joinpath(runner.default_output_file_name).with_suffix(".json")

        expected = "\n".join(json.dumps(result.model_dump_json()) for result in results) + "\n"
        with output_path.open("r") as f:
            assert f.read() == expected

    def test_write_results_as_github_annotation(
            self, runner: ContractsRunner, results: list[Result], faker: Faker, tmp_path: Path
    ):
        output_path = runner.write_results(results, path=tmp_path, output_type="github-annotations")
        assert output_path == tmp_path.joinpath(runner.default_output_file_name).with_suffix(".json")

        assert not any(result.can_format_to_github_annotation for result in results)
        with output_path.open("r") as f:
            assert f.read() == "[]"

        for result in results:
            result.properties_start_line = faker.random_int()
            result.properties_end_line = faker.random_int(min=result.properties_start_line)

        assert all(result.can_format_to_github_annotation for result in results)
        output_path = runner.write_results(results, path=tmp_path, output_type="github-annotations")

        expected = json.dumps([result.as_github_annotation() for result in results], indent=2)
        with output_path.open("r") as f:
            assert f.read() == expected



from pathlib import Path
from typing import Any
from unittest import mock

import pytest
from _pytest.fixtures import FixtureRequest
from dbt.artifacts.resources.v1.components import ColumnInfo
from dbt.artifacts.resources.v1.macro import MacroArgument
from dbt.contracts.graph.nodes import ModelNode, Macro, SourceDefinition

from dbt_contracts.contracts.result import Result, ModelResult, SourceResult, ColumnResult, MacroResult, \
    MacroArgumentResult, RESULT_PROCESSOR_MAP
from dbt_contracts.properties import PropertiesIO
from dbt_contracts.types import ItemT, ParentT


class TestResult:
    @pytest.mark.parametrize("item", ["model", "source", "macro"])
    def test_get_result_type(self, item: str, model: ModelNode, request: FixtureRequest):
        item: ItemT = request.getfixturevalue(item)
        expected = item.resource_type.name.title()
        assert Result._get_result_type(item=item) == expected
        assert Result._get_result_type(item=item, parent=model) == f"Model {expected}"

    # noinspection PyTestUnpassedFixture
    @pytest.fixture
    def result(self, model: ModelNode, source: SourceDefinition) -> Result:
        """Fixture for a Result object"""
        path = PropertiesIO.get_path(source)
        props = {
            "__start_line__": 90,
            "__start_col__": 1,
            "__end_line__": 115,
            "__end_col__": 15,
        }
        properties = PropertiesIO({path: props})

        cls = RESULT_PROCESSOR_MAP[type(model)]
        with (
            mock.patch.object(cls, "_extract_properties_for_item", return_value=props),
            mock.patch.object(PropertiesIO, "__getitem__", return_value=path),
        ):
            result = cls.from_resource(
                item=model,
                parent=source,
                properties=properties,
                result_level="ERROR",
                result_name="Failure",
                message="Something bad happened"
            )

        return result

    def test_from_resource(self, result: Result, model: ModelNode, source: SourceDefinition):
        assert result.name == model.name
        assert result.path == Path(source.original_file_path)
        assert result.result_type == "Source Model"
        assert result.result_level == "ERROR"
        assert result.result_name == "Failure"
        assert result.message == "Something bad happened"
        assert result.properties_path == Path(source.original_file_path)
        assert result.properties_start_line == 90
        assert result.properties_start_col == 1
        assert result.properties_end_line == 115
        assert result.properties_end_col == 15
        assert result.parent_id == source.unique_id
        assert result.parent_name == source.name
        assert result.parent_type == source.resource_type.name.title()

    def test_github_annotation(self, result: Result):
        assert result.can_format_to_github_annotation

        expected_keys = {
            "path",
            "start_line",
            "start_column",
            "end_line",
            "end_column",
            "annotation_level",
            "title",
            "message",
        }
        assert set(result.as_github_annotation()) & expected_keys == expected_keys

        result.properties_start_line = None
        assert not result.can_format_to_github_annotation


class TestModelResult:
    @pytest.fixture
    def properties(self) -> dict[str, Any]:
        """Fixture for a loaded properties object"""
        return {
            "models": [
                {"name": "model1", "config": {"test": 1}},
                {"name": "model2", "config": {"test": 2}},
                {"name": "model3", "config": {"test": 3}},
            ]
        }

    def test_extract_properties_for_item(self, properties: dict[str, Any], model: ModelNode):
        model.name = "does not exist"
        assert ModelResult._extract_properties_for_item(properties, model) is None

        model.name = "model2"
        assert ModelResult._extract_properties_for_item(properties, model) == properties["models"][1]


class TestSourceResult:
    @pytest.fixture
    def properties(self) -> dict[str, Any]:
        """Fixture for a loaded properties object"""
        return {
            "sources": [
                {
                    "name": "source1",
                    "tables": [
                        {"name": "table1", "config": {"test": 1}},
                        {"name": "table2", "config": {"test": 2}},
                        {"name": "table3", "config": {"test": 3}},
                    ],
                },
                {
                    "name": "source2",
                    "tables": [
                        {"name": "other-table1", "config": {"test": 1}},
                        {"name": "other-table2", "config": {"test": 2}},
                        {"name": "other-table3", "config": {"test": 3}},
                    ],
                }
            ]
        }

    def test_extract_properties_for_item(self, properties: dict[str, Any], source: SourceDefinition):
        source.source_name = "does not exist"
        assert SourceResult._extract_properties_for_item(properties, source) is None

        source.source_name = "source1"
        source.name = "table2"
        assert SourceResult._extract_properties_for_item(properties, source) == properties["sources"][0]["tables"][1]


class TestColumnResult:
    @pytest.fixture
    def properties(self) -> dict[str, Any]:
        """Fixture for a loaded properties object"""
        return {
            "models": [
                {"name": "model1", "columns": [{"name": "col1", "tags": ["tag1"]}]},
                {"name": "model2", "columns": [{"name": "col2", "tags": ["tag2"]}]},
                {"name": "model3", "columns": [{"name": "col3", "tags": ["tag3"]}]},
            ],
            "sources": [
                {
                    "name": "source1",
                    "tables": [
                        {"name": "table1", "columns": [{"name": "col1", "tags": ["tag1"]}]},
                        {"name": "table2", "columns": [{"name": "col2", "tags": ["tag2"]}]},
                        {"name": "table3", "columns": [{"name": "col3", "tags": ["tag3"]}]},
                    ],
                },
                {
                    "name": "source2",
                    "tables": [
                        {"name": "other-table1", "columns": [{"name": "col1", "tags": ["tag1"]}]},
                        {"name": "other-table2", "columns": [{"name": "col2", "tags": ["tag2"]}]},
                        {"name": "other-table3", "columns": [{"name": "col3", "tags": ["tag3"]}]},
                    ],
                }
            ]
        }

    def test_extract_properties_for_model(self, properties: dict[str, Any], model: ModelNode, column: ColumnInfo):
        model.name = "model2"
        column.name = "col2"

        expected = properties["models"][1]["columns"][0]
        assert ColumnResult._extract_properties_for_item(properties, item=column, parent=model) == expected

    def test_extract_properties_for_source(
            self, properties: dict[str, Any], source: SourceDefinition, column: ColumnInfo
    ):
        source.source_name = "source1"
        source.name = "table2"
        column.name = "col2"

        expected = properties["sources"][0]["tables"][1]["columns"][0]
        assert ColumnResult._extract_properties_for_item(properties, item=column, parent=source) == expected

    @pytest.mark.parametrize("parent", ["model", "source"])
    def test_get_result_type(self, parent: str, column: ColumnInfo, request: FixtureRequest):
        parent: ParentT = request.getfixturevalue(parent)
        expected = parent.resource_type.name.title()
        assert ColumnResult._get_result_type(item=column, parent=parent) == f"{expected} Column"

    def test_from_resource_gets_index(self, source: SourceDefinition, column: ColumnInfo):
        assert column.name not in ("col1", "col2", "col3")
        source.columns = {
            "col1": ColumnInfo(name="col1", tags=["tag1"]),
            "col2": ColumnInfo(name="col2", tags=["tag2"]),
            "col3": ColumnInfo(name="col3", tags=["tag3"]),
            column.name: column,
        }

        result = ColumnResult.from_resource(
            item=column, parent=source, result_level="ERROR", result_name="Failure", message="Something bad happened"
        )
        assert result.index == 3


class TestMacroResult:
    @pytest.fixture
    def properties(self) -> dict[str, Any]:
        """Fixture for a loaded properties object"""
        return {
            "macros": [
                {"name": "macro1", "arguments": [{"name": "arg1", "description": "I am arg1"}]},
                {"name": "macro2", "arguments": [{"name": "arg2", "description": "I am arg2"}]},
                {"name": "macro3", "arguments": [{"name": "arg3", "description": "I am arg3"}]},
            ]
        }

    def test_extract_properties_for_item(self, properties: dict[str, Any], macro: Macro):
        macro.name = "does not exist"
        assert MacroResult._extract_properties_for_item(properties, item=macro) is None

        macro.name = "macro2"
        assert MacroResult._extract_properties_for_item(properties, item=macro) == properties["macros"][1]


class TestMacroArgumentResult:
    @pytest.fixture
    def properties(self) -> dict[str, Any]:
        """Fixture for a loaded properties object"""
        return {
            "macros": [
                {"name": "macro1", "arguments": [{"name": "arg1", "description": "I am arg1"}]},
                {"name": "macro2", "arguments": [{"name": "arg2", "description": "I am arg2"}]},
                {"name": "macro3", "arguments": [{"name": "arg3", "description": "I am arg3"}]},
            ]
        }

    def test_extract_properties_for_item(self, properties: dict[str, Any], macro: Macro, argument: MacroArgument):
        macro.name = "macro2"
        argument.name = "does not exist"
        assert MacroArgumentResult._extract_properties_for_item(properties, item=argument, parent=macro) is None

        argument.name = "arg2"
        expected = properties["macros"][1]["arguments"][0]
        assert MacroArgumentResult._extract_properties_for_item(properties, item=argument, parent=macro) == expected

    @pytest.mark.parametrize("parent", ["model", "source", "macro"])
    def test_get_result_type(self, parent: str, argument: MacroArgument, request: FixtureRequest):
        parent: ParentT = request.getfixturevalue(parent)
        # always gives the same value
        assert MacroArgumentResult._get_result_type(item=argument, parent=parent) == "Macro Argument"

    def test_from_resource_gets_index(self, macro: Macro, argument: MacroArgument):
        assert argument.name not in ("arg1", "arg2", "arg3")
        macro.arguments = [
            MacroArgument(name="arg1"),
            MacroArgument(name="arg2"),
            argument,
            MacroArgument(name="arg3"),
        ]

        result = MacroArgumentResult.from_resource(
            item=argument, parent=macro, result_level="ERROR", result_name="Failure", message="Something bad happened"
        )
        assert result.index == 2



from abc import ABCMeta, abstractmethod
from collections.abc import Callable, Collection, Iterable
from typing import Any

from dbt_contracts._core import BaseModelConfig
from dbt_contracts.contracts.result import Result

type KeysT[T] = str | Callable[[T], Any]


def get_value_from_object[T](obj: T, key: KeysT[T]) -> Any:
    """
    Get a values from the given `obj` for the given `key`.

    :param obj: The object to get a value from.
    :param key: The key from which to get the value.
        May either be a string of the attribute name, or a lambda function for more complex logic.
    :return: The value from the object.
    """
    return key(obj) if callable(key) else getattr(obj, key)


def get_values_from_object[T](obj: T, keys: Collection[KeysT[T]]) -> Iterable[Any]:
    """
    Get many values from the given `obj` for the given `key`.

    :param obj: The object to get values from.
    :param keys: The keys from which to get the values.
        May either be a collection strings of the attribute name,
        or a collection of lambda functions for more complex logic.
    :return: The value from the object.
    """
    return (get_value_from_object(obj, key) for key in keys)


class ResultsFormatter[T: Result](BaseModelConfig, metaclass=ABCMeta):
    """
    Base class for implementations which format a set of :py:class:`.Result` objects to a string for displaying results.
    Usually used to format results for logging purposes.
    This allows for the separation of how results should be formatted for logging purposes from their implementations.
    """
    @abstractmethod
    def add_results(self, results: Collection[T]) -> None:
        """
        Format the given results and update the stored output to be built.

        :param results: The results to format.
        """
        raise NotImplementedError

    @abstractmethod
    def build(self) -> str:
        """Build the output and return it. This clears the stored output."""
        raise NotImplementedError



from dbt.artifacts.resources.v1.macro import MacroArgument
from dbt.contracts.graph.nodes import Macro

from dbt_contracts.contracts._core import ContractContext
from dbt_contracts.contracts.terms._core import ChildContractTerm, validate_context


class HasType(ChildContractTerm[MacroArgument, Macro]):
    """Check whether macro arguments have a data type configured in their properties."""
    @validate_context
    def run(self, item: MacroArgument, context: ContractContext, parent: Macro = None) -> bool:
        has_type = bool(item.type)
        if not has_type:
            message = "Argument does not have a type configured"
            context.add_result(name=self.name, message=message, item=item, parent=parent)

        return has_type



import pytest

from dbt_contracts.contracts.matchers import RangeMatcher, StringMatcher, PatternMatcher


class TestRangeMatcher:
    def test_checks_range_values_are_valid(self):
        max_min_count_log = "Maximum count must be >= minimum count"
        with pytest.raises(Exception, match=max_min_count_log):
            RangeMatcher(min_count=10, max_count=5)
        with pytest.raises(Exception, match=max_min_count_log):
            RangeMatcher(min_count=12, max_count=2)

    def test_match(self):
        assert RangeMatcher(min_count=1, max_count=10)._match(5, kind="test") is None
        assert RangeMatcher(min_count=5, max_count=5)._match(5, kind="test") is None
        assert RangeMatcher(min_count=4, max_count=6)._match(5, kind="test") is None
        assert RangeMatcher(min_count=6, max_count=6)._match(5, kind="test") is not None
        assert RangeMatcher(min_count=2, max_count=4)._match(5, kind="test") is not None


# noinspection SpellCheckingInspection
class TestStringMatcher:
    def test_match_on_none(self):
        assert StringMatcher()._match(None, None)
        assert not StringMatcher()._match(None, "not none")
        assert not StringMatcher()._match("not none", None)

    def test_match_with_no_processing(self):
        assert not StringMatcher()._match("we are equal", "we are not equal")
        assert StringMatcher()._match("we are equal", "we are equal")

    def test_match_case_insensitive(self):
        assert not StringMatcher(case_insensitive=False)._match("we are equal", "We Are Equal")
        assert StringMatcher(case_insensitive=True)._match("we are equal", "We Are Equal")

    def test_match_ignore_whitespace(self):
        assert not StringMatcher(ignore_whitespace=False)._match("we are equal", "weareequal")
        assert StringMatcher(ignore_whitespace=True)._match("we are equal", "weareequal")

    def test_match_complex(self):
        assert not StringMatcher(compare_start_only=False)._match("we are equal", "we are")
        assert StringMatcher(compare_start_only=True)._match("we are equal", "we are")

        assert StringMatcher(case_insensitive=True, ignore_whitespace=True)._match(
            "we are equal", "WeAreEqual"
        )
        assert StringMatcher(case_insensitive=True, ignore_whitespace=True, compare_start_only=True)._match(
            "we are equal", "WeAre",
        )


class TestPatternMatcher:
    def test_match_on_none_and_no_patterns(self):
        assert not PatternMatcher()._match(None)
        assert not PatternMatcher(include="*")._match(None)
        assert PatternMatcher()._match("i am a value")

    def test_match_includes_all_on_no_include_patterns(self):
        assert PatternMatcher(include=())._match("i am a value")
        assert PatternMatcher(exclude="exclude me")._match("i am a value")

    def test_match_on_simple_patterns(self):
        assert PatternMatcher(include="i am a value")._match("i am a value")
        assert not PatternMatcher(exclude="i am a value")._match("i am a value")

        assert PatternMatcher(include=r".*")._match("i am a value")
        assert not PatternMatcher(exclude=r".*")._match("i am a value")

    def test_match_on_complex_patterns(self):
        assert PatternMatcher(include=[r"i am a \w+", r"i am not a \w+"], match_all=False)._match("i am a value")
        assert not PatternMatcher(include=[r"i am a \w+", r"i am not a \w+"], match_all=True)._match("i am a value")

        assert not PatternMatcher(
            include=[r"i am a \w+", r"i am not a \w+"], exclude=".*value$"
        )._match("i am a value")
        assert PatternMatcher(
            include=r"i am a \w+", exclude=["^this.*" ".*value$"], match_all=True
        )._match("i am a value")
        assert PatternMatcher(
            include=[r"i am a \w+", r"[^\d]+"], exclude=["^this.*" ".*value$"], match_all=True
        )._match("i am a value")



import os
from collections.abc import Collection, Mapping, MutableMapping
from pathlib import Path
from typing import Any

from dbt.artifacts.schemas.catalog import CatalogArtifact
from dbt.contracts.graph.nodes import SourceDefinition
from dbt.flags import get_flags
from dbt_common.contracts.metadata import CatalogTable

from dbt_contracts.types import NodeT


def to_tuple(value: Any) -> tuple:
    """Convert the given value to a tuple"""
    if value is None:
        return tuple()
    elif isinstance(value, tuple):
        return value
    elif isinstance(value, str):
        value = (value,)
    return tuple(value)


def merge_maps[T: MutableMapping](source: T, new: Mapping, extend: bool = True, overwrite: bool = False) -> T:
    """
    Recursively update a given ``source`` map in place with a ``new`` map.

    :param source: The source map.
    :param new: The new map with values to update for the source map.
    :param extend: When a value is a list and a list is already present in the source map, extend the list when True.
        When False, only replace the list if overwrite is True.
    :param overwrite: When True, overwrite any value in the source destructively.
    :return: The updated dict.
    """
    def is_collection(value: Any) -> bool:
        """Return True if ``value`` is of type ``Collection`` and not a string or map"""
        return isinstance(value, Collection) and not isinstance(value, str) and not isinstance(value, Mapping)

    for k, v in new.items():
        if isinstance(v, Mapping) and isinstance(source.get(k, {}), Mapping):
            source[k] = merge_maps(source.get(k, {}), v, extend=extend, overwrite=overwrite)
        elif extend and is_collection(v) and is_collection(source.get(k, [])):
            source[k] = list(to_tuple(source.get(k, []))) + list(to_tuple(v))
        elif overwrite or source.get(k) is None:
            source[k] = v

    return source


def get_matching_catalog_table(item: NodeT, catalog: CatalogArtifact) -> CatalogTable | None:
    """
    Check whether the given `item` exists in the database.

    :param item: The resource to match.
    :param catalog: The catalog of tables.
    :return: The matching catalog table.
    """
    if isinstance(item, SourceDefinition):
        return catalog.sources.get(item.unique_id)
    return catalog.nodes.get(item.unique_id)


def get_absolute_project_path(path: str | Path) -> Path | None:
    """
    Get the absolute path of the given relative `path` in the project directory.
    Only returns the path if it exists.

    :param path: The relative path.
    :return: The absolute project path.
    """
    flags = get_flags()
    project_dir = getattr(flags, "PROJECT_DIR", None) or ""

    if project_dir and (path_in_project := Path(project_dir, path)).exists():
        return path_in_project.resolve()
    elif (path_in_cwd := Path(os.getcwd(), path)).exists():
        return path_in_cwd.resolve()

    raise FileNotFoundError(f"Could not find absolute path for: {path}")



import os
from copy import deepcopy
from pathlib import Path
from random import choice
from unittest import mock

import pytest
from dbt.artifacts.resources import BaseResource
from dbt.artifacts.schemas.catalog import CatalogArtifact
from dbt.contracts.graph.nodes import CompiledNode, ModelNode
from dbt.flags import GLOBAL_FLAGS

from dbt_contracts.contracts.utils import get_matching_catalog_table, get_absolute_project_path, merge_maps


def tests_merge_maps():
    source = {
        1: "value",
        2: {"a": "val a", "b": "val b", "c": {"nested1": "nested val"}},
        3: {"nested1": {"nested2": {"nested3": "old value"}}},
        4: {"a": [1, 2, 3]}
    }
    new = {
        2: {"b": "new value b", "c": {"nested1": "modified nested val"}},
        3: {"nested1": {"nested2": {"nested3": "new value", "new key": "new val"}}},
        4: {"a": [4, 5]}
    }

    test = deepcopy(source)
    merge_maps(source=test, new=new, extend=False, overwrite=False)
    assert test == {
        1: "value",
        2: {"a": "val a", "b": "val b", "c": {"nested1": "nested val"}},
        3: {"nested1": {"nested2": {"nested3": "old value", "new key": "new val"}}},
        4: {"a": [1, 2, 3]}
    }

    test = deepcopy(source)
    merge_maps(source=test, new=new, extend=False, overwrite=True)
    assert test == {
        1: "value",
        2: {"a": "val a", "b": "new value b", "c": {"nested1": "modified nested val"}},
        3: {"nested1": {"nested2": {"nested3": "new value", "new key": "new val"}}},
        4: {"a": [4, 5]}
    }

    test = deepcopy(source)
    merge_maps(source=test, new=new, extend=True, overwrite=False)
    assert test == {
        1: "value",
        2: {"a": "val a", "b": "val b", "c": {"nested1": "nested val"}},
        3: {"nested1": {"nested2": {"nested3": "old value", "new key": "new val"}}},
        4: {"a": [1, 2, 3, 4, 5]}
    }


def test_get_matching_catalog_table(node: CompiledNode, simple_resource: BaseResource, catalog: CatalogArtifact):
    table = get_matching_catalog_table(item=node, catalog=catalog)
    assert table is not None
    # noinspection PyTypeChecker
    assert get_matching_catalog_table(item=simple_resource, catalog=catalog) is None


@pytest.fixture
def relative_path(model: ModelNode, tmp_path: Path) -> Path:
    """Fixture to generate a relative path for testing."""
    paths = [model.original_file_path, model.path]
    if model.patch_path:
        paths.append(model.patch_path.split("://")[1])
    path = choice([path for path in paths if path is not None])

    expected = tmp_path.joinpath(path)
    expected.parent.mkdir(parents=True, exist_ok=True)
    expected.touch()

    return Path(path)


def test_get_absolute_path_in_project_dir(relative_path: Path, tmp_path: Path):
    GLOBAL_FLAGS.PROJECT_DIR = tmp_path
    expected = tmp_path.joinpath(relative_path)
    assert get_absolute_project_path(relative_path) == expected


# TODO: flakey test - fix me
def test_get_absolute_path_in_cwd(relative_path: Path, tmp_path: Path):
    expected = tmp_path.joinpath(relative_path)
    # noinspection SpellCheckingInspection
    with mock.patch.object(os, "getcwd", return_value=str(tmp_path)):
        assert get_absolute_project_path(relative_path) == expected
